import copy
import logging
from typing import Callable, List, Dict

import numpy as np
from torch.utils.data import TensorDataset

from knodle.trainer import AutoTrainer
from knodle.trainer.utils import log_section


logger = logging.getLogger(__name__)


class MultiTrainer:
    """ The factory class for creating training executors
    See See https://medium.com/@geoffreykoh/implementing-the-factory-
    pattern-via-dynamic-registry-and-python-decorators-479fc1537bbe
    """

    registry = {}
    """ Internal registry for available trainers """

    def __init__(self, name: List, **kwargs):
        self.trainer, configs = [], []
        if "trainer_config" in kwargs:
            configs = copy.deepcopy(kwargs["trainer_config"])
        for i, n in enumerate(name):
            if configs:
                kwargs["trainer_config"] = configs[i]
            self.trainer.append(copy.deepcopy(AutoTrainer(n, **kwargs).trainer))

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        for i in range(len(self.trainer)):
            trainer_name = str(type(self.trainer[i])).split(".")[-1][:-2]
            log_section(f"Training {trainer_name}", logger)
            self.trainer[i].train()

    def test(self, test_features: TensorDataset, test_labels: TensorDataset) -> Dict:
        metrics = {}
        for i in range(len(self.trainer)):
            trainer_name = str(type(self.trainer[i])).split(".")[-1][:-2]
            metrics[trainer_name] = self.trainer[i].test(test_features, test_labels)
        return metrics



import logging
from typing import Union, Dict, Iterable

import numpy as np
from scipy import sparse as ss

matrix = Union[np.ndarray, ss.csr_matrix]

logger = logging.getLogger(__name__)


def reduce_rule_matches(
    rule_matches_z: matrix, mapping_rules_labels_t: matrix, drop_rules: bool = False,
    max_rules: int = None, min_coverage: float = None, rule_matches_rest: Dict[str, matrix] = None
) -> Dict[str, matrix]:

    """
    This is the main function to be used for reduction of rules.
    It finds the less effective rules by coverage in the train rule_matches_z and reduces them in the match matrix.
    The reduction is either done by dropping the rules (if drop_rules is True), or by creating one merged rule per label
    from the excessive rules selected for reduction (drop_rules is False).
    The mapping_rules_labels_t matrix is adjusted accordingly. At least one of the criteria [max_rules, min_coverage]
    should be provided.
    Validation and test matches can be provided in rule_matches_rest to be reduced in the same manner as training data.

    Args:
        rule_matches_z: main match matrix, based on which the rules to keep are selected
        mapping_rules_labels_t: rule-to-label mapping, has to be reduced to the selected rules as well
        drop_rules: If True, the rules will be completely discarded, otherwise they will be
            represented as one rule (column) in reduced match matrix.
        max_rules: maximal number of rules to keep as-is. If drop_rules=False, with additional merged rules the total
            number of rules in the output may exceed this limit by #labels at most.
        min_coverage: minimal coverage required for a rule to be kept
        rule_matches_rest: additional rule_matches (dev, test) which should be reduced in the same manner
    Returns: Dictionary with "train_rule_matches_z", "mapping_rules_labels_t" and keys from rule_matches_rest if provided.
    """
    if max_rules is None and min_coverage is None:
        logger.info("No filtering criteria ('max_rule' or 'min_coverage' for rule specified, "
                    "returning the original rule matches.")
        out = {"train_rule_matches_z": rule_matches_z, "mapping_rules_labels_t": mapping_rules_labels_t}
        if rule_matches_rest:
            out.update(rule_matches_rest)
        return out

    rule_kept_ids = _select_rules_for_reduction(rule_matches_z, max_rules, min_coverage)

    # add training matches to the dictionary
    rule_matches_dict = {"train_rule_matches_z": rule_matches_z}
    if rule_matches_rest is not None:
        rule_matches_dict.update(rule_matches_rest)

    if drop_rules:
        return _reduce_by_drop(rule_matches_dict, mapping_rules_labels_t, rule_kept_ids)
    else:
        return _reduce_by_merge(rule_matches_dict, mapping_rules_labels_t, rule_kept_ids)


def _select_rules_for_reduction(
        rule_matches_z: matrix, max_rules: int = None, min_coverage: float = None
) -> np.ndarray:
    """
    Selects rule ids that match the reduction criteria based on rule coverage in the provided match matrix.
    At least one of the criteria [max_rules, min_coverage] should be provided.
    Args:
        rule_matches_z: main match matrix, based on which the rules to keep are selected
        max_rules: maximal number of rules to keep as-is.
        min_coverage: minimal coverage required for a rule to be kept
    Returns: Array with rule ids.
    """
    coverage_per_rule = rule_matches_z.sum(0) / rule_matches_z.shape[0]

    if isinstance(rule_matches_z, ss.csr_matrix):
        # convert from numpy.matrix (result of sparse sum) to numpy.array
        coverage_per_rule = np.array(coverage_per_rule).flatten()

    # create a set to collect ids of rules which will be kept unchanged
    rule_kept_ids = np.array(range(rule_matches_z.shape[1]))

    # take top N rules
    if max_rules is not None:
        max_rule_ids = (coverage_per_rule * -1).argsort(kind="stable")[:max_rules]
        rule_kept_ids = np.intersect1d(rule_kept_ids, max_rule_ids)

    # filter under-coveraged rules
    if min_coverage is not None:
        min_coverage_ids = np.nonzero(coverage_per_rule < min_coverage)[0]
        rule_kept_ids = np.setdiff1d(rule_kept_ids, min_coverage_ids)

    # convert to array to slice matrices
    return rule_kept_ids


def _reduce_by_drop(
        rule_matches_dict: Dict[str, matrix], mapping_rules_labels_t: matrix, rule_kept_ids: np.ndarray
) -> Dict[str, matrix]:
    """
    Drops the rules according to the provided mask. Performs same reduction on all match matrices in the dictionary.
    Corresponding columns of mapping matrix T are dropped as well.
    If N rules were selected to *not* be reduced (there are N entries in the `rule_kept_ids`):
    Args:
        rule_matches_dict: contains the rule match matrices (#objects x N)
        mapping_rules_labels_t: rule-to-class mapping (N x #labels)
        rule_kept_ids: ids of rules that should *not* be reduced (N)

    Returns: Reduced match matrices for all keys in the `rule_matches_dict`, reduced mapping T in a single dictionary.
    """
    output_dict = {}
    for split, match_matrix in rule_matches_dict.items():
        output_dict[split] = match_matrix[:, rule_kept_ids]
    output_dict["mapping_rules_labels_t"] = mapping_rules_labels_t[rule_kept_ids, :]
    return output_dict


def _reduce_by_merge(
        rule_matches_dict: Dict[str, matrix], mapping_rules_labels_t: matrix, rule_kept_ids: np.ndarray
) -> Dict[str, matrix]:
    """
    Leaves rules selected according to the provided mask unchanged
    while merging the others based on their corresponding labels.
    Performs same reduction on all match matrices in the dictionary.
    Corresponding columns of mapping matrix T are processed similarly and will contain an unchanged and merged columns.
    If N rules were selected to *not* be reduced (there are N entries in the `rule_kept_ids`),
    and the other rules correspond to M labels, the resulting set of rules will contain N + M rules,
    Args:
        rule_matches_dict: contains the rule match matrices (#objects x (N + M))
        mapping_rules_labels_t: rule-to-label mapping ((N + M) x #labels)
        rule_kept_ids: ids of rules that should *not* be reduced (N)

    Returns: Reduced match matrices for all keys in the `rule_matches_dict`, reduced mapping T in a single dictionary.
    """

    # get core part of remaining rule matches
    output_dict = _reduce_by_drop(rule_matches_dict, mapping_rules_labels_t, rule_kept_ids)

    # reduce all provided match matrices in the same manner; keep dict keys
    for split, full_match_matrix in rule_matches_dict.items():
        reduced_matches = _get_merged_matrix(
            full_matches=full_match_matrix,
            to_keep_mask=rule_kept_ids,
            label_rule_masks=_get_rule_by_label_iterator(mapping_rules_labels_t)
        )

        # add merged rules to the core matches
        if isinstance(output_dict[split], ss.csr_matrix):
            # add columns with sparse columns, then convert back to rows
            output_dict[split] = ss.hstack([output_dict[split].tocsc(), reduced_matches]).tocsr()
        else:
            output_dict[split] = np.hstack([output_dict[split], reduced_matches])

    # add merged mapping
    merged_mapping = _get_merged_mapping(
        to_keep_ids=rule_kept_ids,
        label_rule_masks=_get_rule_by_label_iterator(mapping_rules_labels_t),
        number_of_labels=mapping_rules_labels_t.shape[1],
        make_sparse=isinstance(mapping_rules_labels_t, ss.csr_matrix)
    )

    if isinstance(merged_mapping, ss.csr_matrix):
        output_dict["mapping_rules_labels_t"] = ss.vstack([output_dict["mapping_rules_labels_t"], merged_mapping])
    else:
        output_dict["mapping_rules_labels_t"] = np.vstack([output_dict["mapping_rules_labels_t"], merged_mapping])

    return output_dict


def _get_rule_by_label_iterator(
    mapping_rules_labels_t: matrix
) -> Iterable[np.ndarray]:
    """
    Get an iterator yielding an rule ids mask for every label, indicating which rules belong to this label.
    """
    for label_id in range(mapping_rules_labels_t.shape[1]):
        column = mapping_rules_labels_t[:, label_id]
        rule_mask = column.nonzero()[0]

        # check if rules for this label also correspond to other labels
        if isinstance(mapping_rules_labels_t, ss.csr_matrix):
            labels_for_rules = mapping_rules_labels_t[rule_mask].sum(0).nonzero()[1].tolist()
        else:
            labels_for_rules = mapping_rules_labels_t[rule_mask].sum(0).nonzero()[0].tolist()

        if labels_for_rules != [label_id]:
            logger.warning(f"Rules for {label_id} point to multiple labels {labels_for_rules}")
        yield rule_mask


def _get_merged_mapping(
    to_keep_ids: np.ndarray, label_rule_masks: Iterable[np.ndarray], number_of_labels: int, make_sparse: bool
) -> np.ndarray:
    """
    Creates a rule-to-label mapping matrix according to the rules selected for reduction.
    Returns: Mapping array of shape M (depends of #labels affected by reduction) x #labels
    """
    mappings_rules_labels = []

    for label_id, label_mask in enumerate(label_rule_masks):
        #to_reduce_label_mask = label_mask & to_reduce_mask
        to_reduce_label_mask = np.setdiff1d(label_mask, to_keep_ids)

        # if there are no rules for this label that should be reduced, skip the label
        if len(to_reduce_label_mask) == 0:
            continue

        mapping_label_row = np.zeros((number_of_labels,))

        # set the new rule to label with label_id
        mapping_label_row[label_id] = 1

        if make_sparse:
            mapping_label_row = ss.csr_matrix(mapping_label_row)

        mappings_rules_labels.append(mapping_label_row)

    if make_sparse:
        return ss.vstack(mappings_rules_labels)
    else:
        return np.vstack(mappings_rules_labels)


def _get_merged_matrix(
    full_matches: matrix, to_keep_mask: np.ndarray, label_rule_masks: Iterable[np.ndarray]
) -> matrix:
    """
    Merge the selected columns from the original match matrix of x rules to a one-per-label rule.
    If a label corresponds to any of the rules selected for reduction, a new merged rule is constructed for this label
    which contains matches from all to-be-reduced rules for this label.

    Args:
        full_matches: original match matrix with all rules: shape #obj x #rules
        to_keep_mask: mask for the rules from 'full_matches' that should be reduced
        label_rule_masks: mapping of label indices to rule masks (shape: #rules)
    Returns: reduced matrix #obj x #labels
    """
    reduced_matches = []

    for label_id, label_mask in enumerate(label_rule_masks):
        # identify relevant rules from all label rules
        to_reduce_label_mask = np.setdiff1d(label_mask, to_keep_mask)

        # if there are no rules for this label that should be reduced, skip the label
        if len(to_reduce_label_mask) == 0:
            continue

        # create a merged match column
        reduced_matches_col = full_matches[:, to_reduce_label_mask].sum(1)

        # take care of multiple matches per row --> reduce them to one
        if isinstance(full_matches, ss.csr_matrix):
            # create sparse *column* for faster column-wise rule match processing
            reduced_matches_col = ss.csc_matrix(reduced_matches_col).minimum(1)
        else:
            reduced_matches_col = np.minimum(reduced_matches_col, 1).reshape(-1, 1)

        reduced_matches.append(reduced_matches_col)

    if isinstance(full_matches, ss.csr_matrix):
        return ss.hstack(reduced_matches)
    else:
        return np.hstack(reduced_matches)



import argparse
import csv
import os
import sys
from pathlib import Path
import logging
from typing import Dict, Tuple, List

from tutorials.data_preprocessing.tac_based_dataset.conll_relation_extraction_dataset.utils import (
    count_file_lines, get_id, update_dict, convert_to_tacred_rel
)

logger = logging.getLogger(__name__)
PRINT_EVERY = 1000000


def reconstruct_arg_pairs_lfs(
        path_data: str,
        path_labels: str,
        path_output: str
) -> None:
    """ Thins function reconstructs argument pairs that have been LFs while constructing the conll noisy dataset """
    Path(path_output).mkdir(parents=True, exist_ok=True)
    labels2ids = get_labels(path_labels)
    rules = get_lfs(path_data, labels2ids)

    with open(os.path.join(path_output, "lfs_no_overlapping_rules.csv"), 'a', newline='') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=["rule", "rule_id", "label", "label_id"])
        writer.writeheader()
        writer = csv.writer(csvfile)
        writer.writerows(rules)


def get_labels(path_labels: str) -> Dict:
    """ Reads the labels from the file and encode them with ids """
    relation2ids = {}
    with open(path_labels, encoding="UTF-8") as file:
        for line in file.readlines():
            relation, relation_enc = line.replace("\n", "").split(",")
            relation2ids[relation] = int(relation_enc)
    return relation2ids


def get_rule(
        subj: list, obj: list, subj_min_token_id: int, obj_min_token_id: int, label, label_id, rule2id, relation2rules, rules
):
    if subj_min_token_id < obj_min_token_id:
        rule = "_".join(subj) + " " + "_".join(obj)
    else:
        rule = "_".join(obj) + " " + "_".join(subj)
    if label_id != "unk":
        rule_id = get_id(rule, rule2id)
        update_dict(label, rule_id, relation2rules)
        rule_info = [str(rule), str(rule_id), str(label), str(label_id)]
        if rule_info not in rules:
            rules.append(rule_info)
    return rules


def extract_subj_obj(
        line: str, subj: list, obj: list, subj_min_token_id: int, obj_min_token_id: int
) -> Tuple[List, List, int, int]:
    splitted_line = line.split("\t")
    token = splitted_line[1]
    if splitted_line[2] == "SUBJECT":
        if not subj_min_token_id:
            subj_min_token_id = int(splitted_line[0])
        subj.append(token)
    elif splitted_line[4] == "OBJECT":
        if not obj_min_token_id:
            obj_min_token_id = int(splitted_line[0])
        obj.append(token)
    return subj, obj, subj_min_token_id, obj_min_token_id


def get_lfs(conll_data: str, labels2ids: Dict) -> list:
    relation2rules, rule2id = {}, {}
    rules = []
    num_lines = count_file_lines(conll_data)
    processed_lines = 0
    with open(conll_data, encoding='utf-8') as f:
        for line in f:
            processed_lines += 1
            line = line.strip()
            if line.startswith("# id="):  # Instance starts
                subj, obj = [], []
                subj_min_token_id, obj_min_token_id = 0, 0
                label, label_id = get_label_n_label_id(line, labels2ids)
            elif line == "":  # Instance ends
                if len(subj) == 0 or len(obj) == 0:
                    continue
                rules = get_rule(
                    subj, obj, subj_min_token_id, obj_min_token_id, label, label_id, rule2id, relation2rules, rules
                )
            elif line.startswith("#"):  # comment
                continue
            else:
                subj, obj, subj_min_token_id, obj_min_token_id = extract_subj_obj(
                    line, subj, obj, subj_min_token_id, obj_min_token_id
                )
            print_progress(processed_lines, num_lines)
    return rules


def get_label_n_label_id(line: str, labels2ids: dict) -> Tuple[str, int]:
    label = convert_to_tacred_rel(line.split(" ")[3][5:])
    if label not in labels2ids and label != "no_relation":
        logger.info(label)
    label_id = encode_labels(label, labels2ids)
    return label, label_id


def print_progress(processed_lines: int, num_lines: int) -> None:
    if processed_lines % (int(round(num_lines / 10))) == 0:
        print(f"Processed {processed_lines / num_lines * 100 :0.0f}%")


def encode_labels(label: str, label2id: dict) -> int:
    """ Encodes labels with corresponding labels id. If relation is unknown, returns special id for unknown relations"""
    return label2id.get(label, "unk")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))
    parser.add_argument("--path_data", help="")
    parser.add_argument("--path_labels", help="")
    parser.add_argument("--path_outputs", help="")

    args = parser.parse_args()
    reconstruct_arg_pairs_lfs(
        args.path_data,
        args.path_labels,
        args.path_outputs
    )



import copy
import logging
import os

import torch
from torch.utils.data import DataLoader
from joblib import dump

from knodle.trainer.baseline.majority import MajorityVoteTrainer
from knodle.trainer.utils import log_section
from knodle.trainer.wscrossweigh.data_splitting_by_rules import k_folds_splitting_by_rules
from knodle.transformation.filter import filter_empty_probabilities
from knodle.transformation.majority import z_t_matrices_to_majority_vote_probs

logger = logging.getLogger(__name__)
torch.set_printoptions(edgeitems=100)


class WSCrossWeighWeightsCalculator(MajorityVoteTrainer):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # save the copy of the original model; later wscrossweigh models for each training with a new hold-out fold
        # will be copied from it
        self.wscrossweigh_model = copy.deepcopy(self.model).to(self.trainer_config.device)
        self.sample_weights = torch.empty(0)

    def calculate_weights(self) -> torch.FloatTensor:
        """
        This function calculates the sample_weights for samples using WSCrossWeigh method
        :return matrix of the sample sample_weights
        """

        # initialize optimizer
        self.trainer_config.optimizer = self.initialise_optimizer()

        if self.trainer_config.folds < 2:
            raise ValueError("Number of folds should be at least 2 to perform WSCrossWeigh denoising")

        logger.info("======= Denoising with WSCrossWeigh is started =======")
        os.makedirs(self.trainer_config.caching_folder, exist_ok=True)

        noisy_y_train = z_t_matrices_to_majority_vote_probs(
            self.rule_matches_z, self.mapping_rules_labels_t, self.trainer_config.other_class_id
        )

        if self.trainer_config.filter_non_labelled:
            self.model_input_x, noisy_y_train, self.rule_matches_z = filter_empty_probabilities(
                self.model_input_x, noisy_y_train, self.rule_matches_z
            )

        # initialise sample weights
        self.sample_weights = self.initialise_sample_weights()

        train_datasets, test_datasets = \
            k_folds_splitting_by_rules(
                self.model_input_x,
                noisy_y_train,
                self.rule_matches_z,
                self.trainer_config.partitions,
                self.trainer_config.folds,
                self.trainer_config.other_class_id
            )

        for iter, (train_dataset, test_dataset) in enumerate(zip(train_datasets, test_datasets)):
            log_section(
                f"WSCrossWeigh Iteration {iter + 1}/{self.trainer_config.partitions * self.trainer_config.folds}:",
                logger
            )

            # for each fold the model is trained from scratch
            self.model = copy.deepcopy(self.wscrossweigh_model).to(self.trainer_config.device)
            test_loader = self._make_dataloader(test_dataset)
            train_loader = self._make_dataloader(train_dataset)
            self._train_loop(train_loader)
            self.cw_test(test_loader)

            log_section(f"WSCrossWeigh Partition {iter + 1} is done", logger)

        dump(self.sample_weights, os.path.join(
            self.trainer_config.caching_folder, f"sample_weights_{self.trainer_config.caching_suffix}.lib"))

        logger.info("======= Denoising with WSCrossWeigh is completed =======")
        return self.sample_weights

    def cw_test(self, test_loader: DataLoader) -> None:
        """
        This function tests of trained WSCrossWeigh model on a hold-out fold, compared the predicted labels with the
        ones got with weak supervision and reduces sample_weights of disagreed samples
        :param test_loader: loader with the data which is used for testing (hold-out fold)
        """
        self.wscrossweigh_model.eval()
        correct_predictions, wrong_predictions = 0, 0

        with torch.no_grad():
            for batch in test_loader:
                features, labels = self._load_batch(batch)
                data_features, data_indices = features[:-1], features[-1]

                outputs = self.wscrossweigh_model(*data_features)
                outputs = outputs[0] if not isinstance(outputs, torch.Tensor) else outputs
                _, predicted = torch.max(outputs.data, -1)
                predictions = predicted.tolist()

                for curr_pred in range(len(predictions)):
                    gold = labels.tolist()[curr_pred]
                    gold_classes = gold.index(max(gold))
                    guess = predictions[curr_pred]
                    if guess != gold_classes:       # todo: what if more than one class could be predicted? e.g. conll
                        wrong_predictions += 1
                        curr_id = data_indices[curr_pred].tolist()
                        self.sample_weights[curr_id] *= self.trainer_config.weight_reducing_rate
                    else:
                        correct_predictions += 1
        logger.info("Correct predictions: {:.3f}%, wrong predictions: {:.3f}%".format(
            correct_predictions * 100 / (correct_predictions + wrong_predictions),
            wrong_predictions * 100 / (correct_predictions + wrong_predictions)))

    def initialise_sample_weights(self) -> torch.FloatTensor:
        """ Initialise a sample_weights matrix (num_samples x 1): weights for all samples equal sample_start_weights """
        return torch.FloatTensor([self.trainer_config.samples_start_weights] * self.model_input_x.tensors[0].shape[0])



import os
import logging
from typing import Union, Dict, Tuple

import skorch
from tqdm.auto import tqdm
from abc import ABC, abstractmethod

import numpy as np
from sklearn.metrics import classification_report

import torch
from torch import Tensor
from torch.nn import Module
from torch.utils.data import TensorDataset, DataLoader
import torch.nn.functional as F

from knodle.evaluation.other_class_metrics import classification_report_other_class
from knodle.transformation.torch_input import input_labels_to_tensordataset, dataset_to_numpy_input
from knodle.transformation.rule_reduction import reduce_rule_matches
from knodle.evaluation.plotting import draw_loss_accuracy_plot

from knodle.trainer.config import TrainerConfig, BaseTrainerConfig
from knodle.trainer.utils.utils import log_section, accuracy_of_probs
from knodle.trainer.utils.checks import check_other_class_id

logger = logging.getLogger(__name__)


class Trainer(ABC):
    def __init__(
            self,
            model: Module,
            mapping_rules_labels_t: np.ndarray,
            model_input_x: TensorDataset,
            rule_matches_z: np.ndarray,
            dev_model_input_x: TensorDataset = None,
            dev_gold_labels_y: TensorDataset = None,
            trainer_config: TrainerConfig = None,
    ):
        """
        Constructor for each Trainer.
            Args:
                model: PyTorch model which will be used for final classification.
                mapping_rules_labels_t: Mapping of rules to labels, binary encoded. Shape: rules x classes
                model_input_x: Input tensors. These tensors will be fed to the provided model.
                rule_matches_z: Binary encoded array of which rules matched. Shape: instances x rules
                trainer_config: Config for different parameters like loss function, optimizer, batch size.
        """
        self.model = model
        self.mapping_rules_labels_t = mapping_rules_labels_t
        self.model_input_x = model_input_x
        self.rule_matches_z = rule_matches_z
        self.dev_model_input_x = dev_model_input_x
        self.dev_gold_labels_y = dev_gold_labels_y

        if trainer_config is None:
            self.trainer_config = TrainerConfig(model)
        else:
            self.trainer_config = trainer_config

    @abstractmethod
    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        pass

    @abstractmethod
    def test(self, test_features: TensorDataset, test_labels: TensorDataset):
        pass

    def initialise_optimizer(self):
        try:
            return self.trainer_config.optimizer(params=self.model.parameters(), lr=self.trainer_config.lr)
        except TypeError:
            logger.info("Wrong optimizer parameters. Optimizer should belong to torch.optim class or be PyTorch "
                        "compatible.")


class BaseTrainer(Trainer):

    def __init__(
            self,
            model: Module,
            mapping_rules_labels_t: np.ndarray,
            model_input_x: TensorDataset,
            rule_matches_z: np.ndarray,
            **kwargs):
        if kwargs.get("trainer_config", None) is None:
            kwargs["trainer_config"] = BaseTrainerConfig()
        super().__init__(model, mapping_rules_labels_t, model_input_x, rule_matches_z, **kwargs)

        check_other_class_id(self.trainer_config, self.mapping_rules_labels_t)

    def _load_train_params(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        if model_input_x is not None and rule_matches_z is not None:
            self.model_input_x = model_input_x
            self.rule_matches_z = rule_matches_z
        if dev_model_input_x is not None and dev_gold_labels_y is not None:
            self.dev_model_input_x = dev_model_input_x
            self.dev_gold_labels_y = dev_gold_labels_y

    def _apply_rule_reduction(self):
        reduced_dict = reduce_rule_matches(
            rule_matches_z=self.rule_matches_z, mapping_rules_labels_t=self.mapping_rules_labels_t,
            drop_rules=self.trainer_config.drop_rules, max_rules=self.trainer_config.max_rules,
            min_coverage=self.trainer_config.min_coverage)
        self.rule_matches_z = reduced_dict["train_rule_matches_z"]
        self.mapping_rules_labels_t = reduced_dict["mapping_rules_labels_t"]

    def _make_dataloader(
            self, dataset: TensorDataset, shuffle: bool = True
    ) -> DataLoader:
        dataloader = DataLoader(
            dataset,
            batch_size=self.trainer_config.batch_size,
            drop_last=False,
            shuffle=shuffle,
        )
        return dataloader

    def _load_batch(self, batch):

        input_batch = [inp.to(self.trainer_config.device) for inp in batch[0: -1]]
        label_batch = batch[-1].to(self.trainer_config.device)

        return input_batch, label_batch

    def _train_loop(
            self, feature_label_dataloader, use_sample_weights: bool = False,
            draw_plot: bool = False
    ):
        log_section("Training starts", logger)

        self.model.to(self.trainer_config.device)
        self.model.train()

        train_losses, train_acc = [], []
        if self.dev_model_input_x is not None:
            dev_losses, dev_acc = [], []

        for current_epoch in range(self.trainer_config.epochs):
            logger.info("Epoch: {}".format(current_epoch))
            epoch_loss, epoch_acc, steps = 0.0, 0.0, 0
            for batch in tqdm(feature_label_dataloader):
                input_batch, label_batch = self._load_batch(batch)
                steps += 1

                if use_sample_weights:
                    input_batch, sample_weights = input_batch[:-1], input_batch[-1]

                # forward pass
                self.trainer_config.optimizer.zero_grad()
                outputs = self.model(*input_batch)
                if isinstance(outputs, torch.Tensor):
                    logits = outputs
                else:
                    logits = outputs[0]

                if use_sample_weights:
                    loss_no_reduction = self.trainer_config.criterion(
                        logits, label_batch, weight=self.trainer_config.class_weights, reduction="none"
                    )
                    loss = (loss_no_reduction * sample_weights).mean()
                else:
                    loss = self.trainer_config.criterion(logits, label_batch, weight=self.trainer_config.class_weights)

                # backward pass
                loss.backward()
                if isinstance(self.trainer_config.grad_clipping, (int, float)):
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.trainer_config.grad_clipping)
                self.trainer_config.optimizer.step()
                acc = accuracy_of_probs(logits, label_batch)

                epoch_loss += loss.detach().item()
                epoch_acc += acc.item()

                # print epoch loss and accuracy after each 10% of training is done
                try:
                    if steps % (int(round(len(feature_label_dataloader) / 10))) == 0:
                        logger.info(f"Train loss: {epoch_loss / steps:.3f}, Train accuracy: {epoch_acc / steps:.3f}")
                except ZeroDivisionError:
                    continue

            avg_loss = epoch_loss / len(feature_label_dataloader)
            avg_acc = epoch_acc / len(feature_label_dataloader)
            train_losses.append(avg_loss)
            train_acc.append(avg_acc)

            logger.info("Epoch train loss: {}".format(avg_loss))
            logger.info("Epoch train accuracy: {}".format(avg_acc))

            if self.dev_model_input_x is not None:
                dev_clf_report, dev_loss = self.test(
                    self.dev_model_input_x, self.dev_gold_labels_y, loss_calculation=True)
                dev_losses.append(dev_loss)
                dev_acc.append(dev_clf_report["accuracy"])
                logger.info("Epoch development accuracy: {}".format(dev_clf_report["accuracy"]))

            # saving model
            if self.trainer_config.saved_models_dir is not None:
                model_path = os.path.join(
                    self.trainer_config.saved_models_dir,
                    f"model_state_dict_epoch_{current_epoch}.pt"
                )
                torch.save(self.model.cpu().state_dict(), model_path)
                self.model.to(self.trainer_config.device)

        log_section("Training done", logger)

        if draw_plot:
            if self.dev_model_input_x:
                draw_loss_accuracy_plot(
                    {"train loss": train_losses, "train acc": train_acc, "dev loss": dev_losses, "dev acc": dev_acc}
                )
            else:
                draw_loss_accuracy_plot({"train loss": train_losses, "train acc": train_acc})

        self.model.eval()


    def _prediction_loop(
            self, feature_label_dataloader: DataLoader, loss_calculation: str = False
    ) -> [np.ndarray, np.ndarray]:

        self.model.to(self.trainer_config.device)
        self.model.eval()
        predictions_list, label_list = [], []
        dev_loss, dev_acc = 0.0, 0.0

        i = 0
        # Loop over predictions
        with torch.no_grad():
            for batch in tqdm(feature_label_dataloader):
                input_batch, label_batch = self._load_batch(batch)

                # forward pass
                self.trainer_config.optimizer.zero_grad()
                outputs = self.model(*input_batch)
                if isinstance(outputs, torch.Tensor):
                    prediction_vals = outputs
                else:
                    prediction_vals = outputs[0]

                if loss_calculation:
                    dev_loss += self._calculate_dev_loss(prediction_vals, label_batch.long())

                # add predictions and labels
                predictions = np.argmax(prediction_vals.detach().cpu().numpy(), axis=-1)
                predictions_list.append(predictions)
                label_list.append(label_batch.detach().cpu().numpy())

        predictions = np.squeeze(np.hstack(predictions_list))
        gold_labels = np.squeeze(np.hstack(label_list))

        return predictions, gold_labels, dev_loss

    def _calculate_dev_loss(self, predictions: Tensor, labels: Tensor) -> Tensor:
        """ Calculates the loss on the dev set using given criterion"""
        predictions_one_hot = F.one_hot(predictions.argmax(1), num_classes=self.trainer_config.output_classes).float()
        labels_one_hot = F.one_hot(labels, self.trainer_config.output_classes)
        loss = self.trainer_config.criterion(predictions_one_hot, labels_one_hot)
        return loss.detach()

    def test(
            self, features_dataset: TensorDataset, labels: TensorDataset, loss_calculation: bool = False
    ) -> Tuple[Dict, Union[float, None]]:

        gold_labels = labels.tensors[0].cpu().numpy()

        if isinstance(self.model, skorch.NeuralNetClassifier):
            # when the pytorch model is wrapped as a sklearn model (e.g. cleanlab)
            predictions = self.model.predict(dataset_to_numpy_input(features_dataset))
        else:
            feature_label_dataset = input_labels_to_tensordataset(features_dataset, gold_labels)
            feature_label_dataloader = self._make_dataloader(feature_label_dataset, shuffle=False)
            predictions, gold_labels, dev_loss = self._prediction_loop(feature_label_dataloader, loss_calculation)

        if self.trainer_config.evaluate_with_other_class:
            clf_report = classification_report_other_class(
                y_true=gold_labels, y_pred=predictions, ids2labels=self.trainer_config.ids2labels,
                other_class_id=self.trainer_config.other_class_id
            )
        else:
            clf_report = classification_report(y_true=gold_labels, y_pred=predictions, output_dict=True)

        if loss_calculation:
            return clf_report, dev_loss / len(feature_label_dataloader)
        else:
            return clf_report, None



from typing import Tuple, Union, List

import numpy as np
from torch.utils.data import TensorDataset


def filter_tensor_dataset_by_indices(dataset: TensorDataset, filter_ids: Union[np.ndarray, List[int]]) -> TensorDataset:
    """Filters each tensor of a TensorDataset, given some "filter_ids".

    Args:
        dataset: TensorDataset with a list of tensors, each having first dimension N
        filter_ids: A list of K indices to be kept, K <= N

    Returns: TensorDataset with filtered indices
    """
    new_tensors = []
    for i in range(len(dataset.tensors)):
        new_tensors.append(dataset.tensors[i][filter_ids])
    dataset_new = TensorDataset(*new_tensors)

    return dataset_new


def filter_empty_probabilities(
        input_data_x: TensorDataset, class_probas_y: np.ndarray, rule_matches_z: np.ndarray = None
) -> Union[Tuple[TensorDataset, np.ndarray, np.ndarray], Tuple[TensorDataset, np.ndarray]]:
    """Delete rows of TensorDataset's where the cumulative probability equals 0.
    Args:
        input_data_x: A TensorDataset serving as input to a model
        class_probas_y: Array, holding class probabilities, shape=num_samples, num_classes
        rule_matches_z: optional array with rules matched in samples. If given, will also be filtered
    :return: Modified TensorDataset's
    """
    if len(class_probas_y.shape) != 2:
        raise ValueError("y_probs needs to be a matrix of dimensions num_samples x num_classes")

    prob_sums = class_probas_y.sum(axis=-1)
    non_zeros = np.where(prob_sums != 0)[0]

    new_x = filter_tensor_dataset_by_indices(dataset=input_data_x, filter_ids=non_zeros)

    if rule_matches_z is not None:
        return new_x, class_probas_y[non_zeros], rule_matches_z[non_zeros]

    return new_x, class_probas_y[non_zeros]


def filter_probability_threshold(
        input_data_x: TensorDataset, class_probas_y: np.ndarray, rule_matches_z: np.ndarray = None,
        probability_threshold: float = 0.7
) -> Union[Tuple[TensorDataset, np.ndarray, np.ndarray], Tuple[TensorDataset, np.ndarray]]:
    """Filters instances where no single class probability exceeds "probability_threshold".
    """
    prob_sums = class_probas_y.max(axis=-1)
    conclusive_idx = np.where(prob_sums >= probability_threshold)[0]

    new_x = filter_tensor_dataset_by_indices(dataset=input_data_x, filter_ids=conclusive_idx)

    if rule_matches_z is not None:
        return new_x, class_probas_y[conclusive_idx], rule_matches_z[conclusive_idx]

    return new_x, class_probas_y[conclusive_idx]



import logging

from numpy import ndarray
from knodle.trainer.config import BaseTrainerConfig


def check_other_class_id(trainer_config: BaseTrainerConfig, mapping_rules_labels_t: ndarray):
    # check and derive other_class_id from class mappings if neccessary
    if trainer_config.other_class_id is None:
        if not trainer_config.filter_non_labelled:
            trainer_config.other_class_id = mapping_rules_labels_t.shape[1]
    elif trainer_config.other_class_id < 0:
        raise RuntimeError("Label for negative samples should be greater than 0 for correct matrix multiplication")
    elif trainer_config.other_class_id < mapping_rules_labels_t.shape[1] - 1:
        logging.warning(f"Negative class {trainer_config.other_class_id} is already present in data")



import numpy as np
import torch
from torch.utils.data import TensorDataset


def input_labels_to_tensordataset(model_input_x: TensorDataset, labels: np.ndarray) -> TensorDataset:
    """
    This function takes Dataset with data features (num_samples x features dimension x features) and
    labels (num_samples x labels dimension) and turns it into one Dataset
    """
    model_tensors = model_input_x.tensors
    input_label_dataset = TensorDataset(*model_tensors, torch.from_numpy(labels))

    return input_label_dataset


def input_info_labels_to_tensordataset(
        model_input_x: TensorDataset, input_info: np.ndarray, labels: np.ndarray
) -> TensorDataset:
    """
    This function takes Dataset with data features (num_samples x features dimension x features),
    labels (num_samples x labels dimension) and some additional information about data encoded as
    a numpy array (num_samples x n; could be sample weights, sample indices etc) and turns it into one Dataset
    """
    model_tensors = model_input_x.tensors
    input_ids_label_dataset = TensorDataset(*model_tensors, torch.from_numpy(input_info), torch.from_numpy(labels))

    return input_ids_label_dataset


def dataset_to_numpy_input(model_input_x: TensorDataset) -> np.ndarray:
    if len(model_input_x.tensors) == 1:
        return model_input_x.tensors[0].detach().cpu().numpy()
    else:
        raise ValueError(f"Selected denoising method accepts input features encoded with one tensor only, while "
                         f"{len(model_input_x.tensors) + 1} input tensors were given. Please use another input "
                         f"encoding or another denoising method.")



import copy

import numpy as np
from sklearn.base import RegressorMixin
from tqdm import tqdm

from knodle.trainer.wscrossweigh.data_splitting_by_rules import (
    k_folds_splitting_by_rules, k_folds_splitting_by_signatures
)


def estimate_cv_predicted_probabilities_split_by_rules(
        model_input_x: np.ndarray,
        labels: np.ndarray,
        rule_matches_z: np.ndarray,
        model: RegressorMixin,
        num_classes: int,
        cv_n_folds: int = 5,
        seed: int = None,
        other_class_id: int = None
):
    cv_train_datasets, cv_holdout_datasets = k_folds_splitting_by_rules(
        model_input_x, labels, rule_matches_z, partitions=1, num_folds=cv_n_folds, seed=seed,
        other_class_id=other_class_id
    )

    return compute_psx_matrix(model, cv_train_datasets, cv_holdout_datasets, labels, num_classes)


def estimate_cv_predicted_probabilities_split_by_signatures(
        model_input_x: np.ndarray,
        labels: np.ndarray,
        rule_matches_z: np.ndarray,
        model: RegressorMixin,
        num_classes: int,
        cv_n_folds: int = 5,
        seed: int = None,
        other_class_id: int = None
):

    cv_train_datasets, cv_holdout_datasets = k_folds_splitting_by_signatures(
        model_input_x, labels, rule_matches_z, partitions=1, num_folds=cv_n_folds, seed=seed,
        other_class_id=other_class_id
    )

    return compute_psx_matrix(model, cv_train_datasets, cv_holdout_datasets, labels, num_classes)


def compute_psx_matrix(model, cv_train_datasets, cv_holdout_datasets, labels, num_classes):

    psx = np.zeros((len(labels), num_classes))

    for k, (cv_train_dataset, cv_holdout_dataset) in tqdm(enumerate(zip(cv_train_datasets, cv_holdout_datasets))):
        model_copy = copy.deepcopy(model)
        X_train_cv, y_train_cv = cv_train_dataset.tensors[0].cpu().detach().numpy(), \
                                 cv_train_dataset.tensors[1].cpu().detach().numpy()
        X_holdout_cv, indices_holdout_cv, y_holdout_cv = cv_holdout_dataset.tensors[0].cpu().detach().numpy(), \
                                                         cv_holdout_dataset.tensors[1].cpu().detach().numpy(), \
                                                         cv_holdout_dataset.tensors[2].cpu().detach().numpy()

        # y_train_cv = np.argmax(y_train_cv, axis=1)
        model_copy.fit(X_train_cv, y_train_cv)
        psx_cv = model_copy.predict_proba(X_holdout_cv)  # P(s = k|x) # [:,1]
        psx[indices_holdout_cv] = psx_cv

    return psx



from typing import Dict

import numpy as np
from sklearn.metrics.classification import classification_report

from knodle.transformation.majority import probabilies_to_majority_vote, z_t_matrices_to_majority_vote_probs


def majority_sklearn_report(
        rule_matches_z: np.array, mapping_rules_labels_t: np.array, labels_y: np.array
) -> Dict:
    rule_counts_probs = z_t_matrices_to_majority_vote_probs(rule_matches_z, mapping_rules_labels_t)

    kwargs = {"choose_random_label": True}
    majority_y = np.apply_along_axis(probabilies_to_majority_vote, axis=1, arr=rule_counts_probs, **kwargs)

    sklearn_report = classification_report(labels_y, majority_y, output_dict=True)

    return sklearn_report


def sklearn_report_to_knodle_report(sklearn_report: Dict, prefix: str = None):
    if prefix is None:
        report = {
            f"accuracy": sklearn_report["accuracy"],
            f"macro_f1": sklearn_report["macro avg"]["f1-score"],
            f"weighted_f1": sklearn_report["weighted avg"]["f1-score"],
        }
    else:
        report = {
            f"{prefix}accuracy": sklearn_report["accuracy"],
            f"{prefix}macro_f1": sklearn_report["macro avg"]["f1-score"],
            f"{prefix}weighted_f1": sklearn_report["weighted avg"]["f1-score"],
        }
    return report



from knodle.trainer.config import TrainerConfig

from knodle.trainer.baseline.majority import MajorityVoteTrainer, MajorityConfig
from knodle.trainer.knn_aggregation.knn import KNNAggregationTrainer, KNNConfig
from knodle.trainer.snorkel.snorkel import SnorkelKNNAggregationTrainer, SnorkelTrainer, SnorkelConfig, SnorkelKNNConfig
from knodle.trainer.wscrossweigh.wscrossweigh import WSCrossWeighTrainer, WSCrossWeighConfig

from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.auto_config import AutoConfig


import numpy as np
import scipy.sparse as sp
import warnings

from torch.utils.data import TensorDataset

from knodle.transformation.filter import filter_empty_probabilities, filter_probability_threshold


def probabilies_to_majority_vote(
        probs: np.ndarray, choose_random_label: bool = True, other_class_id: int = None
) -> int:
    """Transforms a vector of probabilities to its majority vote. If there is one class with clear majority, return it.
    If there are more than one class with equal probabilities: either select one of the classes randomly or assign to
    the sample the other class id.
 
    Args:
        probs: Vector of probabilities for 1 sample. Shape: classes x 1
        choose_random_label: Choose a random label, if there's no clear majority.
        other_class_id: Class ID being used, if there's no clear majority
    Returns: An array of classes.
    """
    if choose_random_label and other_class_id is not None:
        raise ValueError("You can either choose a random class, or transform undefined cases to an other class.")

    row_max = np.max(probs)
    num_occurrences = (row_max == probs).sum()
    if num_occurrences == 1:
        return int(np.argmax(probs))
    elif choose_random_label:
        max_ids = np.where(probs == row_max)[0]
        return int(np.random.choice(max_ids))
    elif other_class_id is not None:
        return other_class_id
    else:
        raise ValueError("Specify a way how to resolve unclear majority votes.")


def z_t_matrices_to_majority_vote_probs(
        rule_matches_z: np.ndarray, mapping_rules_labels_t: np.ndarray, other_class_id: int = None
) -> np.ndarray:
    """
    This function calculates a majority vote probability for all rule_matches_z. The difference from simple
    get_majority_vote_probs function is the following: samples, where no rules matched (that is, all elements in
    the corresponding raw in rule_matches_z matrix equal 0), are assigned to no_match_class (that is, a value in the
    corresponding column in rule_counts_probs matrix is changed to 1).

    Args:
        rule_matches_z: Binary encoded array of which rules matched. Shape: instances x rules
        mapping_rules_labels_t: Mapping of rules to labels, binary encoded. Shape: rules x classes
        other_class_id: Class which is chosen, if no function is hitting.
    Returns: Array with majority vote probabilities. Shape: instances x classes
    """

    if rule_matches_z.shape[1] != mapping_rules_labels_t.shape[0]:
        raise ValueError(f"Dimensions mismatch! Z matrix has shape {rule_matches_z.shape}, while "
                         f"T matrix has shape {mapping_rules_labels_t.shape}")

    if isinstance(rule_matches_z, sp.csr_matrix):
        rule_counts = rule_matches_z.dot(mapping_rules_labels_t)
        if isinstance(rule_counts, sp.csr_matrix):
            rule_counts = rule_counts.toarray()
    else:
        rule_counts = np.matmul(rule_matches_z, mapping_rules_labels_t)

    if other_class_id:
        if other_class_id < 0:
            raise RuntimeError("Label for negative samples should be greater than 0 for correct matrix multiplication")
        if other_class_id < mapping_rules_labels_t.shape[1] - 1:
            warnings.warn(f"Negative class {other_class_id} is already present in data")
        if rule_counts.shape[1] == other_class_id:
            rule_counts = np.hstack((rule_counts, np.zeros([rule_counts.shape[0], 1])))
            rule_counts[~rule_counts.any(axis=1), other_class_id] = 1
        elif rule_counts.shape[1] >= other_class_id:
            rule_counts[~rule_counts.any(axis=1), other_class_id] = 1
        else:
            raise ValueError("Other class id is incorrect")
    rule_counts_probs = rule_counts / rule_counts.sum(axis=1).reshape(-1, 1)
    rule_counts_probs[np.isnan(rule_counts_probs)] = 0
    return rule_counts_probs


def z_t_matrices_to_majority_vote_labels(
        rule_matches_z: np.ndarray, mapping_rules_labels_t: np.ndarray,
        choose_random_label: bool = True, other_class_id: int = None
) -> np.array:
    """Computes the majority labels. If no clear "winner" is found, other_class_id is used instead.
    Args:
        rule_matches_z: Binary encoded array of which rules matched. Shape: instances x rules
        mapping_rules_labels_t: Mapping of rules to labels, binary encoded. Shape: rules x classes
        choose_random_label: Whether a random label is chosen, if there's no clear majority vote.
        other_class_id: the id of other class, i.e. the class of negative samples
    Returns: Decision per sample. Shape: (instances, )
    """
    rule_counts_probs = z_t_matrices_to_majority_vote_probs(rule_matches_z, mapping_rules_labels_t)

    kwargs = {"choose_random_label": choose_random_label, "other_class_id": other_class_id}
    majority_labels = np.apply_along_axis(probabilies_to_majority_vote, axis=1, arr=rule_counts_probs, **kwargs)
    return majority_labels


def input_to_majority_vote_input(
        rule_matches_z: np.ndarray,
        mapping_rules_labels_t: np.ndarray,
        model_input_x: TensorDataset,
        use_probabilistic_labels: bool = True,
        filter_non_labelled: bool = True,
        probability_threshold: int = None,
        other_class_id: int = None,
) -> np.ndarray:
    """
    This function calculates noisy labels y_hat from Knodle Z and T matrices.
    :param model_input_x:
    :param rule_matches_z: binary encoded array of which rules matched. Shape: instances x rules
    :param mapping_rules_labels_t: mapping of rules to labels, binary encoded. Shape: rules x classes
    :param filter_non_labelled: boolean value, whether the no matched samples should be filtered out or not.
    :param other_class_id: the id of other class, i.e. the class of no matched samples, if they are to be stored.
    :param use_probabilistic_labels: boolean value, whether the output labels should be in form of probabilistic labels
    or single values.
    :return:
    """
    if other_class_id is not None and filter_non_labelled:
        raise ValueError("You can either filter samples with no weak labels or add them to the other class.")

    noisy_y_train = z_t_matrices_to_majority_vote_probs(rule_matches_z, mapping_rules_labels_t, other_class_id)

    if filter_non_labelled and probability_threshold is not None:
        raise ValueError("You can either filter all non labeled samples or those that have probabilities below "
                         "some threshold.")

    #  filter out samples where no pattern matched
    if filter_non_labelled:
        model_input_x, noisy_y_train, rule_matches_z = filter_empty_probabilities(
            model_input_x, noisy_y_train, rule_matches_z
        )

    #  filter out samples where that have probabilities below the threshold
    elif probability_threshold is not None:
        model_input_x, noisy_y_train = filter_probability_threshold(
            model_input_x, noisy_y_train, probability_threshold=probability_threshold
            )

    if not use_probabilistic_labels:
        # convert labels represented as a prob distribution to a single label using majority voting
        kwargs = {"choose_random_label": True, "other_class_id": other_class_id}
        noisy_y_train = np.apply_along_axis(probabilies_to_majority_vote, axis=1, arr=noisy_y_train, **kwargs)

    return model_input_x, noisy_y_train, rule_matches_z



import matplotlib.pyplot as plt


def draw_loss_accuracy_plot(curves: dict) -> None:
    """ The function creates a plot of 4 curves and displays it"""
    colors = "bgrcmyk"
    color_index = 0
    epochs = range(1, len(next(iter(curves.values()))) + 1)

    for label, value in curves.items():
        plt.plot(epochs, value, c=colors[color_index], label=label)
        color_index += 1

    plt.xticks(epochs)
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)
    plt.show()



import numpy as np

from knodle.transformation.rule_label_format import transform_snorkel_matrix_to_z_t


def test_transform_snorkel_matrix_to_z_t():

    lambda_matrix = np.array([[-1, 2, -1, 1], [1, -1, -1, 1], [1, 2, 0, -1]])

    correct_z_matrix = np.array([[0, 1, 0, 1], [1, 0, 0, 1], [1, 1, 1, 0]])

    correct_t_matrix = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0], [0, 1, 0]])

    z, t = transform_snorkel_matrix_to_z_t(lambda_matrix)

    np.array_equal(correct_z_matrix, z)
    np.array_equal(correct_t_matrix, t)



from torch.nn import CrossEntropyLoss

from tests.trainer.generic import std_trainer_input_2

from knodle.trainer.cleanlab.cleanlab import CleanLabTrainer
from knodle.trainer.cleanlab.config import CleanLabConfig


def test_cleanlab_base_test(std_trainer_input_2):
    (
        model,
        inputs_x, mapping_rules_labels_t, train_rule_matches_z,
        test_dataset, test_labels
    ) = std_trainer_input_2

    config = CleanLabConfig(cv_n_folds=2, criterion=CrossEntropyLoss, use_probabilistic_labels=False)

    trainer = CleanLabTrainer(
        model=model,
        mapping_rules_labels_t=mapping_rules_labels_t,
        model_input_x=inputs_x,
        rule_matches_z=train_rule_matches_z,
        trainer_config=config
    )

    trainer.train()
    clf_report, _ = trainer.test(test_dataset, test_labels)

    # Check that this runs without error
    assert True



#!/usr/bin/env python
# coding: utf-8

"""
IMDB Dataset - Create Weak Supervision Sources and Get the Weak Data Annotations

This notebook shows how to use keywords as a weak supervision source on the example of a well-known IMDB Movie Review dataset, which targets a binary sentiment analysis task.

The original dataset has gold labels, but we will use these labels only for evaluation purposes, since we want to test models under the weak supervision setting with Knodle. The idea behind it is that you don't have a dataset which is purely labeled with strong supervision (manual) and instead use heuristics (e.g. rules) to obtain a weak labeling. In the following tutorial, we will look for certain keywords expressing positive and negative sentiments that can be helpful to distinguish between classes. Specifically, we use the [Opinion lexicon](https://www.cs.uic.edu/~liub/FBS/sentiment-analysis.html) of the University of Illinois at Chicago.

First, we load the dataset from a Knodle dataset collection. Then, we will create [Snorkel](https://www.snorkel.org/) labeling functions from two sets of keywords and apply them to the IMDB reviews. Please keep in mind, that keyword matching can be done without Snorkel; however, we enjoy the transparent annotation functionality of this library in our tutorial. 
Each labeling function (i.e. keyword) will be further associated with a respective target label. This concludes the annotation step.

To estimate how good our weak labeling works on its own, we will use the resulting keyword matches together with a basic majority vote model. Finally, the preprocessed data will be saved in a knodle-friendly format, so that other denoising models can be trained with the IMDB dataset.
The IMDB dataset available in the Knodle collection was downloaded from [Kaggle](https://www.kaggle.com/lakshmi25npathi/imdb-dataset-of-50k-movie-reviews) in January 2021. 
"""

import os
from joblib import dump
from tqdm import tqdm

import pandas as pd 
import numpy as np 
from scipy.sparse import csr_matrix

from bs4 import BeautifulSoup
from snorkel.labeling import LabelingFunction, PandasLFApplier, LFAnalysis

from knodle.transformation.rule_label_format import transform_snorkel_matrix_to_z_t
from knodle.transformation.majority import z_t_matrices_to_majority_vote_labels

# client to access the dataset collection
from minio import Minio
client = Minio("knodle.dm.univie.ac.at", secure=False)

# init pandas parameters
tqdm.pandas()
pd.set_option('display.max_colwidth', -1)

# Constants for Snorkel labeling functions
POSITIVE = 1
NEGATIVE = 0
ABSTAIN = -1
COLUMN_WITH_TEXT = "reviews_preprocessed"


## Download the dataset
data_path = "../../../data_from_minio/imdb"

# Together with the IMDB data, let us also collect the keywords that we will need later.
files = [("IMDB Dataset.csv",), ("keywords", "negative-words.txt"), ("keywords", "positive-words.txt")]

for file in tqdm(files):
    client.fget_object(
        bucket_name="knodle",
        object_name=os.path.join("datasets/imdb", *file),
        file_path=os.path.join(data_path, file[-1]),
    )

imdb_dataset_raw = pd.read_csv(os.path.join(data_path, "IMDB Dataset.csv"))
print("Gold label counts")
print(imdb_dataset_raw.groupby("sentiment").count())

# ## Preprocess dataset

# The dataset contains many HTML tags. We'll remove them
def strip_html(text):
    soup = BeautifulSoup(text, "html.parser")
    return soup.get_text()

imdb_dataset_raw[COLUMN_WITH_TEXT] = imdb_dataset_raw[COLUMN_WITH_TEXT].apply(
    lambda x: strip_html(x))

# ## Keywords
# 
# For weak supervision sources we use sentiment keyword lists for positive and negative words.
# https://www.cs.uic.edu/~liub/FBS/sentiment-analysis.html
# 
# We have downloaded them from the Knodle collection earlier, with the IMDB dataset. 
# 
# After parsing the keywords from separate files, they are stored in a pd.DataFrame with the corresponding sentiment as "label".

positive_keywords = pd.read_csv(os.path.join(data_path, "positive-words.txt"), sep=" ", header=None, error_bad_lines=False, skiprows=30)
positive_keywords.columns = ["keyword"]
positive_keywords["label"] = "positive"

negative_keywords = pd.read_csv(os.path.join(data_path, "negative-words.txt"),
                                sep=" ", header=None, error_bad_lines=False,  encoding='latin-1', skiprows=30)
negative_keywords.columns = ["keyword"]
negative_keywords["label"] = "negative"

all_keywords = pd.concat([positive_keywords, negative_keywords])
all_keywords.label.value_counts()

# remove overlap of keywords between two sentiments
all_keywords.drop_duplicates('keyword', inplace=True)


# ## Labeling Functions
# 
# Now we start to build labeling functions with Snorkel with these keywords and check the coverage.
# 
# This is an iterative process of course so we surely have to add more keywords and regulary expressions ;-) 

def keyword_lookup(x, keyword, label):
    return label if keyword in x[COLUMN_WITH_TEXT].lower() else ABSTAIN

def make_keyword_lf(keyword: str, label: str) -> LabelingFunction:
    """
    Creates labeling function based on keyword.
    Args:
        keyword: what keyword should be look for
        label: what label does this keyword imply

    Returns: LabelingFunction object

    """
    return LabelingFunction(
        name=f"keyword_{keyword}",
        f=keyword_lookup,
        resources=dict(keyword=keyword, label=label),
    )

def create_labeling_functions(keywords: pd.DataFrame) -> np.ndarray:
    """
    Create Labeling Functions based on the columns keyword and regex. Appends column lf to df.

    Args:
        keywords: DataFrame with processed keywords

    Returns:
        All labeling functions. 1d Array with shape: (number_of_lfs x 1)
    """
    keywords = keywords.assign(lf=keywords.progress_apply(
        lambda x:make_keyword_lf(x.keyword, x.label_id), axis=1
    ))
    lfs = keywords.lf.values
    return lfs

all_keywords["label_id"] = all_keywords.label.map({'positive':POSITIVE, 'negative':NEGATIVE})
labeling_functions = create_labeling_functions(all_keywords)


# ### Apply Labeling Functions
# We get a matrix with all labeling functions applied. This matrix has the shape $(instances \times labeling functions)$

applier = PandasLFApplier(lfs=labeling_functions)
applied_lfs = applier.apply(df=imdb_dataset_raw)

print("Shape of applied labeling functions: ", applied_lfs.shape)
print("Number of reviews", len(imdb_dataset_raw))
print("Number of labeling functions", len(labeling_functions))


# ### Analysis 
# 
# Now we can analyse some basic stats about our labeling functions. The main figures are:
# 
# - Coverage: How many labeling functions match at all
# - Overlaps: How many labeling functions overlap with each other (e.g. awesome and amazing)
# - Conflicts: How many labeling functions overlap and have different labels (e.g. awesome and bad)
# - Correct: Correct LFs
# - Incorrect: Incorrect Lfs

lf_analysis = LFAnalysis(L=applied_lfs, lfs=labeling_functions).lf_summary()
print("LF analysis:")
print(lf_analysis)
print("LF analysis mean values:")
print(pd.DataFrame(lf_analysis.mean()))
print("LF analysis median values:")
print(pd.DataFrame(lf_analysis.median()))

# Lets have a look at some examples that were labeled by a positive keyword. You can see, that the true label for some of them is negative.
# consider 1110th keyword
kw = all_keywords.iloc[1110]
print(f"1110th keyword is '{kw.keyword}' and labels with '{kw.label}'")
print()

# sample 2 random examples where the 50th LF assigned a positive label
examples = imdb_dataset_raw.iloc[applied_lfs[:, 1110] == POSITIVE].loc[:2, [COLUMN_WITH_TEXT,'sentiment']]
print("This keyword has matched the following instances with labels:")
print()

for idx, ex in examples.iterrows():
    print("{} - {}".format(ex.sentiment, ex[COLUMN_WITH_TEXT]))
    print()

    
# ## Transform rule matches
# 
# To work with knodle the dataset needs to be transformed into a binary array $Z$
# 
# (shape: `#instances x # rules_`).
# 
# Where a cell $Z_{ij}$ means that for instance i
# 
#     0 -> Rule j didn't match 
#     1 -> Rule j matched
# 
# Furthermore, we need a matrix `mapping_rule_labels` which has a mapping of all rules to labels, stored in a binary manner as well 
# 
# (shape `#rules x #labels`).

rule_matches, mapping_rules_labels = transform_snorkel_matrix_to_z_t(applied_lfs)


# ### Majority Vote
# 
# Now we make a majority vote based on all rule matches. First we get the `rule_counts` by multiplying `rule_matches` with the `mapping_rules_labels`, then we divide it sumwise by the sum to get a probability value. In the end we counteract the divide with zero issue by setting all nan values to zero. All this happens in the `z_t_matrices_to_majority_vote_labels` function.

# the ties are resolved randomly internally, so the predictions might slightly vary
pred_labels = z_t_matrices_to_majority_vote_labels(rule_matches, mapping_rules_labels)

# There are more positive labels predicted by the majority vote.
labels, counts = np.unique(pred_labels, return_counts=True)

# accuracy of the weak labels
acc = (pred_labels == imdb_dataset_raw.sentiment.map({'positive':POSITIVE, 'negative':NEGATIVE})).mean()
print(f"Accuracy of majority voting: {acc}")

      
# ## Save Files
rule_matches_sparse = csr_matrix(rule_matches)
dump(rule_matches_sparse, os.path.join(data_path, "rule_matches.lib"))
dump(mapping_rules_labels,  os.path.join(data_path, "mapping_rules_labels.lib"))

# We also save the preprocessed texts with labels to use them later to evalutate a classifier.
imdb_dataset_raw['label_id'] = imdb_dataset_raw.sentiment.map({'positive':POSITIVE, 'negative':NEGATIVE})
imdb_dataset_raw.to_csv(os.path.join(data_path, 'imdb_data_preprocessed.csv'), index=None)

all_keywords.to_csv(os.path.join(data_path, 'keywords.csv'), index=None)

# Now, we have created a weak supervision dataset. Of course it is not perfect but it is something with which we can compare performances of different denoising methods with. :-) 



import os
from typing import List

from tqdm.auto import tqdm

import joblib
from minio import Minio

import pandas as pd
import numpy as np
import scipy.sparse as sp

import torch
from torch.utils.data import TensorDataset
from transformers import AutoTokenizer, AutoModelForSequenceClassification, AdamW

from examples.trainer.preprocessing import convert_text_to_transformer_input
from knodle.trainer import AutoTrainer, AutoConfig


# This python script contains rarely any explanation. For more description, we refer to the corresponding
# jupyter notebook. There the steps are explained in more detail


# Define some functions


def np_array_to_tensor_dataset(x: np.ndarray) -> TensorDataset:
    """

    :rtype: object
    """
    if isinstance(x, sp.csr_matrix):
        x = x.toarray()
    x = torch.from_numpy(x)
    x = TensorDataset(x)
    return x


# Define constants
imdb_data_dir = os.path.join(os.getcwd(), "datasets", "spouse")
processed_data_dir = os.path.join(imdb_data_dir, "processed")
os.makedirs(processed_data_dir, exist_ok=True)

# Download data
client = Minio("knodle.dm.univie.ac.at", secure=False)
files = [
    "df_train.csv", "df_dev.csv", "df_test.csv",
    "train_rule_matches_z.lib", "dev_rule_matches_z.lib", "test_rule_matches_z.lib",
    "mapping_rules_labels_t.lib"
]
for file in tqdm(files):
    client.fget_object(
        bucket_name="knodle",
        object_name=os.path.join("datasets/spouse/processed", file),
        file_path=os.path.join(processed_data_dir, file),
    )

# Load data into memory
df_train = pd.read_csv(os.path.join(processed_data_dir, "df_train.csv"))
df_dev = pd.read_csv(os.path.join(processed_data_dir, "df_dev.csv"))
df_test = pd.read_csv(os.path.join(processed_data_dir, "df_test.csv"))

mapping_rules_labels_t = joblib.load(os.path.join(processed_data_dir, "mapping_rules_labels_t.lib"))

train_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "train_rule_matches_z.lib"))
dev_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "dev_rule_matches_z.lib"))
test_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "test_rule_matches_z.lib"))

print(f"Train Z dimension: {train_rule_matches_z.shape}")
print(f"Train avg. matches per sample: {train_rule_matches_z.sum() / train_rule_matches_z.shape[0]}")

model_name = "distilbert-base-uncased"
tokenizer = AutoTokenizer.from_pretrained(model_name)

X_train = convert_text_to_transformer_input(tokenizer, df_train["sample"].tolist())
X_dev = convert_text_to_transformer_input(tokenizer, df_dev["sample"].tolist())
X_test = convert_text_to_transformer_input(tokenizer, df_test["sample"].tolist())

y_dev = np_array_to_tensor_dataset(df_dev['label'].values)
y_test = np_array_to_tensor_dataset(df_test['label'].values)

# Load AutoTrainer

model = AutoModelForSequenceClassification.from_pretrained(model_name)

trainer_type = "majority"
custom_model_config = AutoConfig.create_config(
    name=trainer_type,
    optimizer=AdamW,
    lr=1e-4,
    batch_size=16,
    epochs=2,
    filter_non_labelled=True
)

print(custom_model_config)

trainer = AutoTrainer(
    name="majority",
    model=model,
    mapping_rules_labels_t=mapping_rules_labels_t,
    model_input_x=X_train,
    rule_matches_z=train_rule_matches_z,
    dev_model_input_x=X_dev,
    dev_gold_labels_y=y_dev,
    trainer_config=custom_model_config,
)

# Run training
trainer.train()

# Run evaluation
eval_dict, _ = trainer.test(X_test, y_test)
print(f"Accuracy: {eval_dict.get('accuracy')}")



import torch
from torch.utils.data import TensorDataset
import pytest
import numpy as np

from knodle.trainer.wscrossweigh.data_splitting_by_rules import get_rules_sample_ids, get_samples_labels_idx_by_rule_id


@pytest.fixture(scope='session')
def get_test_data():

    rule_assignments_t = np.array([[1, 0, 0],
                                   [1, 0, 0],
                                   [0, 1, 0],
                                   [0, 1, 0],
                                   [0, 0, 1],
                                   [0, 0, 1],
                                   [0, 0, 1]])

    inputs_x = TensorDataset(torch.Tensor(np.array([[0, 0, 0, 0, 0, 0, 0, 0],
                                                    [1, 1, 1, 1, 1, 1, 1, 1],
                                                    [2, 2, 2, 2, 2, 2, 2, 2],
                                                    [3, 3, 3, 3, 3, 3, 3, 3],
                                                    [4, 4, 4, 4, 4, 4, 4, 4],
                                                    [5, 5, 5, 5, 5, 5, 5, 5],
                                                    [6, 6, 6, 6, 6, 6, 6, 6],
                                                    [7, 7, 7, 7, 7, 7, 7, 7]])))

    rule_matches_z = np.array([[0, 1, 1, 0, 0, 0, 0],
                               [1, 0, 1, 1, 0, 0, 0],
                               [0, 0, 0, 0, 1, 0, 0],
                               [0, 0, 0, 0, 0, 0, 1],
                               [0, 0, 0, 0, 0, 1, 0],
                               [1, 0, 1, 0, 0, 0, 0],
                               [1, 1, 1, 1, 0, 0, 0],
                               [0, 0, 0, 1, 0, 0, 0]])

    rule2sample_id = {0: {1, 5, 6},
                              1: {0, 6},
                              2: {0, 1, 5, 6},
                              3: {1, 6, 7},
                              4: {2},  # no rel sample
                              5: {4},  # no rel sample
                              6: {3}}  # no rel sample

    return inputs_x, rule_matches_z, rule_assignments_t, rule2sample_id


@pytest.fixture(scope='session')
def get_cw_data_test(get_test_data):
    # no filtering, no no_match samples
    test_rules_idx = [1, 2]
    labels = np.array([[0.5, 0.5, 0], [0.3, 0.7, 0], [0, 0, 1], [0, 0, 1],
                       [0, 0, 1], [0.5, 0.5, 0], [0.5, 0.5, 0], [0, 1, 0]])

    return [[get_test_data[0],
             labels,
             test_rules_idx,
             get_test_data[3],
             torch.Tensor(np.array([[0, 0, 0, 0, 0, 0, 0, 0],
                                    [1, 1, 1, 1, 1, 1, 1, 1],
                                    [5, 5, 5, 5, 5, 5, 5, 5],
                                    [6, 6, 6, 6, 6, 6, 6, 6]])),
             np.array([[0.5, 0.5, 0], [0.3, 0.7, 0], [0.5, 0.5, 0], [0.5, 0.5, 0]]),
             np.array([0, 1, 5, 6])
             ]]


@pytest.fixture(scope='session')
def get_cw_data_train(get_test_data):
    # rules to be filtered: 1, 2, no_match samples should be added

    test_samples_idx = [0, 6, 1, 5]
    train_rules_idx = [0, 3, 4, 5, 6]  # 1, 5, 6, 7 --> delete inters --> 7 --> + no rel --> 2, 3, 4, 7
    labels = np.array([[0.5, 0.5, 0], [0.3, 0.7, 0], [0, 0, 1], [0, 0, 1],
                       [0, 0, 1], [0.5, 0.5, 0], [0.5, 0.5, 0], [0, 1, 0]])

    return [[get_test_data[0],
             labels,
             train_rules_idx,
             get_test_data[3],
             test_samples_idx,
             torch.Tensor(np.array([[2, 2, 2, 2, 2, 2, 2, 2],
                                    [3, 3, 3, 3, 3, 3, 3, 3],
                                    [4, 4, 4, 4, 4, 4, 4, 4],
                                    [7, 7, 7, 7, 7, 7, 7, 7]])),
             np.array([[0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 1, 0]]),
             np.array([2, 3, 4, 7])
             ]]


def test_sample_ids_matched_rules_correspondence(get_test_data):
    assert get_rules_sample_ids(get_test_data[1]) == get_test_data[3]


def test_get_cw_data_test(get_cw_data_test):
    for data in get_cw_data_test:
        samples, labels, ids = get_samples_labels_idx_by_rule_id(
            data[0], data[1], data[2], data[3]
        )

        assert torch.equal(samples.tensors[0], data[4])
        np.testing.assert_array_equal(labels, data[5])
        np.testing.assert_array_equal(ids, data[6])


def test_get_cw_data_train(get_cw_data_train):
    for data in get_cw_data_train:
        samples, labels, ids = get_samples_labels_idx_by_rule_id(
            data[0], data[1], data[2], data[3], data[4]
        )

        assert torch.equal(samples.tensors[0], data[5])
        np.testing.assert_array_equal(labels, data[6])
        np.testing.assert_array_equal(ids, data[7])


def test_random_check(get_cw_data_train):
    for data in get_cw_data_train:
        samples, labels, ids = get_samples_labels_idx_by_rule_id(
            data[0], data[1], data[2], data[3], data[4]
        )
        rnd_tst = np.random.randint(0, samples.tensors[0].shape[0])  # take some random index
        tst_sample = samples.tensors[0][rnd_tst, :]
        tst_idx = ids[rnd_tst]
        tst_label = labels[rnd_tst, :] if len(labels.shape) > 1 else labels[rnd_tst]

        tst_sample_true = data[5][rnd_tst, :]
        tst_label_true = data[6][rnd_tst, :] if len(data[6].shape) > 1 else data[6][rnd_tst]
        tst_idx_true = data[7][rnd_tst]

        assert torch.equal(tst_sample, tst_sample_true)
        np.testing.assert_array_equal(tst_label, tst_label_true)
        np.testing.assert_array_equal(tst_idx, tst_idx_true)



import argparse
import os
import sys
from typing import Dict

import numpy as np
import pandas as pd
import torch
from torch import Tensor, LongTensor
from torch.optim import Adam
from torch.utils.data import TensorDataset

from knodle.evaluation.other_class_metrics import score
from knodle.model.bidirectional_lstm_model import BidirectionalLSTM
from knodle.trainer.wscrossweigh.config import WSCrossWeighConfig
from knodle.trainer.wscrossweigh.wscrossweigh import WSCrossWeighTrainer
from examples.utils import read_train_dev_test, get_samples_list

NUM_CLASSES = 42
MAXLEN = 50
SPECIAL_TOKENS = ["<PAD>", "<UNK>"]
CLASS_WEIGHTS = torch.FloatTensor([2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                                   2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                                   2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 1.0])


def train_wscrossweigh(
        path_to_data: str,
        path_labels: str,
        path_emb: str,
        path_sample_weights: str = None,
) -> None:
    """ Training the BiLSTM model with WSCrossWeigh denoising algorithm """

    labels2ids = read_labels_from_file(path_labels, "no_relation")
    word2id, word_embedding_matrix = vocab_and_vectors(path_emb)

    df_train, df_dev, df_test, z_train_rule_matches, z_test_rule_matches, t_mapping_rules_labels = \
        read_train_dev_test(path_to_data, if_dev_data=True)

    train_input_x = get_samples_features(df_train, word2id, samples_column_num=1)

    dev_dataset = get_samples_features(df_dev, word2id, samples_column_num=1)
    dev_labels_dataset = TensorDataset(LongTensor(list(df_dev.iloc[:, 4])))

    test_dataset = get_samples_features(df_test, word2id, samples_column_num=1)
    test_labels_dataset = TensorDataset(LongTensor(list(df_test.iloc[:, 4])))

    os.makedirs(path_sample_weights, exist_ok=True)

    parameters = {
        "lr": 1e-4,
        "cw_lr": 0.8,
        "epochs": 5,
        "cw_partitions": 2,
        "cw_folds": 5,
        "cw_epochs": 2,
        "weight_rr": 0.7,
        "samples_start_weights": 4.0
    }

    model = BidirectionalLSTM(
        word_embedding_matrix.shape[0], word_embedding_matrix.shape[1], word_embedding_matrix, NUM_CLASSES
    )

    custom_wscrossweigh_config = WSCrossWeighConfig(
        output_classes=NUM_CLASSES,
        class_weights=CLASS_WEIGHTS,
        filter_non_labelled=True,
        if_set_seed=True,
        epochs=parameters.get("epochs"),
        batch_size=16,
        optimizer=Adam,
        lr=parameters.get("lr"),
        grad_clipping=5,
        partitions=parameters.get("cw_partitions"),
        folds=parameters.get("cw_folds"),
        weight_reducing_rate=parameters.get("weight_rr"),
        samples_start_weights=parameters.get("samples_start_weights")
    )

    trainer = WSCrossWeighTrainer(
        model=model,
        mapping_rules_labels_t=t_mapping_rules_labels,
        model_input_x=train_input_x,
        dev_model_input_x=dev_dataset,
        dev_gold_labels_y=dev_labels_dataset,
        rule_matches_z=z_train_rule_matches,
        trainer_config=custom_wscrossweigh_config,
        evaluation_method="tacred",
        dev_labels_ids=labels2ids,
        use_weights=True,
        run_classifier=True
    )
    trainer.train()
    clf_report, _ = trainer.test(test_dataset, test_labels_dataset)
    print(clf_report)


def get_samples_features(input_data: pd.DataFrame, word2id: dict, samples_column_num: int = None) -> TensorDataset:
    """ Encodes input samples with glove vectors and returns as a Dataset """
    enc_input_samples = encode_samples(get_samples_list(input_data, samples_column_num), word2id, MAXLEN)
    inputs_x_tensor = torch.LongTensor(enc_input_samples)
    inputs_x_dataset = torch.utils.data.TensorDataset(inputs_x_tensor)
    return inputs_x_dataset


def read_labels_from_file(path_labels: str, negative_label: str) -> dict:
    """ Reads the labels from the file and encode them with ids """
    relation2ids = {}
    with open(path_labels, encoding="UTF-8") as file:
        for line in file.readlines():
            relation, relation_enc = line.replace("\n", "").split(",")
            relation2ids[relation] = int(relation_enc)

    # add no_match label
    if negative_label:
        relation2ids[negative_label] = max(list(relation2ids.values())) + 1

    return relation2ids


def encode_samples(raw_samples: list, word2id: dict, maxlen: int) -> list:
    """ This function turns raw text samples into encoded ones using the given word2id dict """
    enc_input_samples = []
    for sample in raw_samples:
        enc_tokens = [word2id.get(token, 1) for token in sample.lstrip().split(" ")]
        enc_input_samples.append(np.asarray(add_padding(enc_tokens, maxlen), dtype="float32"))
    return enc_input_samples


def add_padding(tokens: list, maxlen: int) -> list:
    """ Provide padding of the encoded tokens to the maxlen; if length of tokens > maxlen, reduce it to maxlen """
    padded_tokens = [0] * maxlen
    for token in range(0, min(len(tokens), maxlen)):
        padded_tokens[token] = tokens[token]
    return padded_tokens


def vocab_and_vectors(filename: str) -> (dict, np.ndarray):
    """
    Reads pretrained word embedding and builds 1) a matrix (words x embedding dim), 2) word to id dictionary
    :param filename: path to file with pretrained word embeddings
    :return: word2id, embedding matrix
    """
    with open(filename, encoding="UTF-8") as in_file:
        parts = in_file.readline().strip().split(" ")
        word_embedding_matrix = np.zeros(
            (int(parts[0]) + len(SPECIAL_TOKENS), int(parts[1]))
        )
        word2id = dict()
        for idx, token in enumerate(SPECIAL_TOKENS):
            word2id[token] = idx
        nextword_id = len(SPECIAL_TOKENS)
        for line in in_file:
            parts = line.strip().split(" ")
            word = parts[0]
            if word not in word2id:
                emb = [float(v) for v in parts[1:]]
                word_embedding_matrix[nextword_id] = emb
                word2id[word] = nextword_id
                nextword_id += 1
    return word2id, word_embedding_matrix


# deprecated
def test(model, trainer, test_features: TensorDataset, test_labels: Tensor, labels2ids: Dict) -> Dict:
    feature_labels_dataset = TensorDataset(test_features.tensors[0], test_labels)
    feature_labels_dataloader = trainer._make_dataloader(feature_labels_dataset)
    model.eval()
    all_predictions, all_labels = torch.Tensor(), torch.Tensor()
    for features, labels in feature_labels_dataloader:
        outputs = model(features)
        _, predicted = torch.max(outputs, 1)
        all_predictions = torch.cat([all_predictions, predicted])
        all_labels = torch.cat([all_labels, labels])
    predictions_idx, test_labels_idx = (all_predictions.detach().type(torch.IntTensor).tolist(),
                                        all_labels.detach().type(torch.IntTensor).tolist())
    idx2labels = dict([(value, key) for key, value in labels2ids.items()])
    predictions = [idx2labels[p] for p in predictions_idx]
    test_labels = [idx2labels[p] for p in test_labels_idx]
    return score(test_labels, predictions, verbose=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))
    parser.add_argument("--path_to_data", help="Path to the folder where all input files are stored.")
    parser.add_argument("--path_to_label_ids", help="Path to the file information about labels and their ids is stored")
    parser.add_argument("--path_to_word_embeddings", help="Path to the file with pretrained Glove embeddings used for "
                                                          "samples encoding")
    parser.add_argument("--sample_weights", help="Path to the folder that either sample weights will be saved to"
                                                 "or will be loaded from")
    parser.add_argument("--num_classes", help="Number of classes")

    args = parser.parse_args()

    train_wscrossweigh(
        args.path_to_data, args.path_labels, args.path_emb, args.sample_weights
    )



import os

from knodle.trainer.baseline.config import MajorityConfig
from knodle.trainer.auto_config import AutoConfig


@AutoConfig.register("knn")
class KNNConfig(MajorityConfig):
    def __init__(
            self,
            k: int = None,
            radius: float = None,
            use_approximation: bool = False,
            activate_no_match_instances: bool = True,
            n_jobs_for_index: int = 4,
            **kwargs
    ):
        """
        A default configuration of KNNConfig Trainer.

        :param k: number of neighbors that are to be found
        :param radius: a radius of a point or points the neighbors are to be found within
        :param use_approximation: if set to True, the approximated ANN will be used instead of kNN
        :param activate_no_match_instances: if set to True, the zero-match rows for knn construction & activation will
        be ignored
        :param n_jobs_for_index:

        """
        super().__init__(**kwargs)
        self.k = k
        self.radius = radius
        self.use_approximation = use_approximation
        self.activate_no_match_instances = activate_no_match_instances
        self.n_jobs = n_jobs_for_index

        if self.k is not None and self.radius is not None:
            raise RuntimeError(
                "The kNN trainer can either use the radius or the number of "
                "neighbours to denoise by neighborhood activation"
            )

        if self.k is None and self.use_approximation:
            raise RuntimeError(
                "The kNN trainer can only use the radius for exact neighbor search "
                "Distance-based selection is currently unavailable for approximate NN."
            )

        # Currently impossible, though can (should?) be done in the future.
        if not self.use_approximation and not self.activate_no_match_instances:
            raise RuntimeError(
                "The kNN trainer with exact neighbor selection always uses all of the instances. "
                "Either 'activate_no_match_instances' or 'use_approximation' has to be set to True."
            )

        # if both number of neighbors and radius size are empty, set k=2 by default
        if self.k is None and self.radius is None:
            self.k = 2

    def get_cache_file(self):
        nn_type = "ann" if self.use_approximation else "knn"
        file_tags = f"{self.k}_{nn_type}"
        if self.caching_suffix:
            file_tags += f"_{self.caching_suffix}"

        cache_file = os.path.join(
            self.caching_folder,
            f"denoised_rule_matches_z_{file_tags}.lib"
        )

        return cache_file



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from knodle.trainer.utils.denoise import activate_neighbors


def test_denoise_knn():
    ## case 1 ##
    test_array = np.array([[0, 1], [1, 0]])
    right_result = np.array([[1, 1], [1, 1]])
    indices = np.array([[0, 1], [1, 0]])

    denoised_z = activate_neighbors(test_array, indices)
    assert_array_equal(denoised_z, right_result)

    # make sure all old matches persist and have same match values (currently always 1)
    assert denoised_z[test_array == 1].sum() == test_array[test_array == 1].sum()

    ## case 2 ##
    test_array = np.diag(np.ones((3,)))
    indices = np.array([
        np.array([0, 1, 2]),
        np.array([0]),
        np.array([1])
    ])

    right_result = np.diag(np.ones((3,)))
    right_result[0, :] = 1
    right_result[1, 0] = 1
    right_result[2, 1] = 1

    denoised_z = activate_neighbors(test_array, indices)
    assert_array_equal(denoised_z, right_result)
    assert denoised_z[test_array == 1].sum() == test_array[test_array == 1].sum()

    ## case 3 ##
    test_array = np.diag(np.ones((3,)))
    indices = np.array([
        np.array([0, 1]),
        np.array([1, 2]),
        np.array([0, 2])
    ])

    right_result = np.ones((3, 3))
    right_result[0, 2] = 0
    right_result[1, 0] = 0
    right_result[2, 1] = 0

    denoised_z = activate_neighbors(test_array, indices)
    assert_array_equal(denoised_z, right_result)
    assert denoised_z[test_array == 1].sum() == test_array[test_array == 1].sum()



import pytest

import numpy as np

from knodle.transformation.majority import (
    probabilies_to_majority_vote, z_t_matrices_to_majority_vote_probs, z_t_matrices_to_majority_vote_labels
)


def test_probabilies_to_majority_vote_fixed():
    # format: (probabilities, gold_result, settings)
    probs_gold_result_settings = [
        (np.array([0.5, 0.2, 0.3]), 0, {"choose_random_label": False, "other_class_id": -1}),
        (np.array([0.5, 0.2, 0.3]), 0, {"choose_random_label": True, "other_class_id": None}),
        (np.array([0.5, 0.2, 0.3]), 0, {"choose_random_label": False, "other_class_id": -1}),
        (np.array([0.5, 0.2, 0.3]), 0, {"choose_random_label": False, "other_class_id": None}),

        (np.array([0.5, 0.5, 0.0]), -1, {"choose_random_label": None, "other_class_id": -1}),
        (np.array([0.0, 0.0, 0.0]), -1, {"choose_random_label": None, "other_class_id": -1})
    ]

    for probs, gold_label, settings in probs_gold_result_settings:
        result = probabilies_to_majority_vote(probs, **settings)

        assert isinstance(result, int)
        assert result == gold_label


def test_probabilies_to_majority_vote_random():
    # format: (probabilities, gold_result, settings)
    probs_gold_result_settings = [
        (np.array([0.5, 0.5, 0.0]), [0, 1], {"choose_random_label": True, "other_class_id": None}),

        (np.array([0.3, 0.3, 0, 2, 0.3]), [0, 1, 3], {"choose_random_label": True, "other_class_id": None}),

        (np.array([0.0, 0.0, 0.0]), [-1], {"choose_random_label": False, "other_class_id": -1}),
        (np.array([0.0, 0.0, 0.0]), [0, 1, 2], {"choose_random_label": True, "other_class_id": None})
    ]

    for probs, gold_label, settings in probs_gold_result_settings:
        result = probabilies_to_majority_vote(probs, **settings)

        assert isinstance(result, int)
        assert result in gold_label


def test_probabilies_to_majority_vote_errors():
    probs, gold_label, settings = (
        np.array([0.0, 0.0, 0.0]), [], {"choose_random_label": False, "other_class_id": None}
    )

    with pytest.raises(ValueError):
        result = probabilies_to_majority_vote(probs, **settings)


@pytest.fixture
def prob_values():
    z = np.zeros((4, 4))
    t = np.zeros((4, 2))

    z[0, 0] = 1
    z[0, 2] = 1
    z[1, 0] = 1
    z[1, 1] = 1
    z[1, 2] = 1
    z[1, 3] = 1
    z[2, 1] = 1

    t[0, 0] = 1
    t[1, 1] = 1
    t[2, 1] = 1
    t[3, 1] = 1

    gold_probs = np.array([
        [0.5, 0.5],
        [0.25, 0.75],
        [0, 1],
        [0, 0]
    ])

    gold_labels = np.array([-1, 1, 1, -1])

    return z, t, gold_probs, gold_labels


def test_get_majority_vote_labels(prob_values):
    z, t, _, gold_labels = prob_values

    majority_labels = z_t_matrices_to_majority_vote_labels(z, t, choose_random_label=False, other_class_id=-1)
    print(majority_labels)
    assert np.array_equal(majority_labels, gold_labels)


def test_get_majority_vote_probs(prob_values):
    z, t, gold_probs, _ = prob_values

    majority_probs = z_t_matrices_to_majority_vote_probs(z, t)
    assert np.array_equal(gold_probs, majority_probs)


import numpy as np
import torch
from torch.utils.data import TensorDataset

from knodle.model.logistic_regression_model import LogisticRegressionModel
from knodle.trainer.baseline.majority import MajorityVoteTrainer


def test_train():
    num_samples = 64
    num_features = 16
    num_rules = 6
    num_classes = 2

    model = LogisticRegressionModel(num_features, num_classes)

    x_np = np.ones((num_samples, num_features)).astype(np.float32)
    x_tensor = torch.from_numpy(x_np)
    model_input_x = TensorDataset(x_tensor)

    rule_matches_z = np.zeros((num_samples, num_rules))
    rule_matches_z[0, 0] = 1
    rule_matches_z[1:, 1] = 1

    mapping_rules_labels_t = np.zeros((num_rules, num_classes))
    mapping_rules_labels_t[:, 1] = 1

    trainer = MajorityVoteTrainer(
        model=model,
        mapping_rules_labels_t=mapping_rules_labels_t,
        model_input_x=model_input_x,
        rule_matches_z=rule_matches_z,
    )

    trainer.train()

    y_np = np.ones((num_samples,))
    y_labels = TensorDataset(torch.from_numpy(y_np))
    metrics, _ = trainer.test(model_input_x, y_labels)

    # Check whether the code ran up to here
    assert 2 == 2



import pathlib
from typing import Callable, Dict
import os
import logging

from snorkel.classification import cross_entropy_with_probs

import torch
from torch import Tensor
from torch.optim import SGD
from torch.optim.optimizer import Optimizer

from knodle.trainer.utils.utils import check_and_return_device, set_seed

logger = logging.getLogger(__name__)


class TrainerConfig:
    def __init__(
            self,
            criterion: Callable[[Tensor, Tensor], float] = cross_entropy_with_probs,
            batch_size: int = 32,
            optimizer: Optimizer = None,
            lr: int = 0.01,
            output_classes: int = 2,
            class_weights: Tensor = None,
            epochs: int = 3,
            seed: int = None,
            grad_clipping: int = None,
            device: str = None,
            caching_folder: str = os.path.join(pathlib.Path().absolute(), "cache"),
            caching_suffix: str = "",
            saved_models_dir: str = None
    ):
        """
        A default and minimum sufficient configuration of a Trainer instance.

        :param criterion: a usual PyTorch criterion; computes a gradient according to a given loss function
        :param batch_size: a usual PyTorch batch_size; the number of training examples utilized in one training iteration
        :param optimizer: a usual PyTorch optimizer; which is used to solve optimization problems by minimizing the
        function
        :param lr: a usual PyTorch learning rate; tuning parameter in an optimization algorithm that determines the step
        size at each iteration while moving toward a minimum of a loss function
        :param output_classes: the number of classes used in classification
        :param class_weights: introduce the weight of each class. By default, all classes have the same weights 1.0.
        :param epochs: the number of epochs the classification model will be trained
        :param seed: the desired seed for generating random numbers
        :param grad_clipping: if set to True, gradient norm of an iterable of parameters will be clipped
        :param device: what device the model will be trained on (CPU/CUDA)
        :param caching_folder: a path to the folder where cache will be saved (default: root/cache)
        :param caching_suffix: a specific index that could be added to the caching file name (e.g. in WSCrossWeigh for
        sample weights calculated in different iterations and stored in different files.)
        :param saved_models_dir: a path to the folder where trained models will be stored. If None, the trained models
        won't be stored.
        """
        self.seed = seed
        if self.seed is not None:
            set_seed(seed)

        self.caching_suffix = caching_suffix
        self.caching_folder = caching_folder
        os.makedirs(self.caching_folder, exist_ok=True)
        logger.info(f"The cache will be saved to {self.caching_folder} folder")

        # create directory where saved models will be stored
        if saved_models_dir:
            self.saved_models_dir = saved_models_dir
            os.makedirs(self.saved_models_dir, exist_ok=True)
        else:
            self.saved_models_dir = caching_folder
        logger.info(f"The trained models will be saved to the {self.saved_models_dir} directory.")

        self.criterion = criterion
        self.lr = lr
        self.batch_size = batch_size
        self.output_classes = output_classes
        self.grad_clipping = grad_clipping
        self.device = torch.device(device) if device is not None else check_and_return_device()
        logger.info(f"Model will be trained on {self.device}")

        if epochs <= 0:
            raise ValueError("Epochs needs to be positive")
        self.epochs = epochs

        if optimizer is None:
            logger.info(f"Defaulting to SGD optimizer as none specified in the config.")
            self.optimizer = SGD
        else:
            self.optimizer = optimizer

        if class_weights is None:
            self.class_weights = torch.tensor([1.0] * self.output_classes)
        else:
            if len(class_weights) != self.output_classes:
                raise Exception("Wrong class sample_weights initialisation!")
            self.class_weights = class_weights


class BaseTrainerConfig(TrainerConfig):
    def __init__(
            self,
            filter_non_labelled: bool = True,
            other_class_id: int = None,
            evaluate_with_other_class: bool = False,
            ids2labels: Dict = None,
            max_rules: int = None,
            min_coverage: float = None,
            drop_rules: bool = False,
            **kwargs
    ):
        """
        Additionally provided parameters needed for handling the cases where there are data samples with no rule
        matched (filtering OR introducing the other class + training & evaluation with other class).

        :param filter_non_labelled: if True, the samples with no rule matched will be filtered out from the dataset
        :param other_class_id: id of the negative class; if set, the samples with no rule matched will be assigned to it
        :param evaluate_with_other_class: if set to True, the evaluation will be done with respect to the negative class
        (for more details please see knodle/evaluation/other_class_metrics.py file)
        :param ids2labels: dictionary {label id: label}, which is needed to perform evaluation with the negative class
        """
        super().__init__(**kwargs)
        self.filter_non_labelled = filter_non_labelled
        self.other_class_id = other_class_id
        self.evaluate_with_other_class = evaluate_with_other_class
        self.ids2labels = ids2labels

        if self.other_class_id is not None and self.filter_non_labelled:
            raise ValueError("You can either filter samples with no weak labels or add them to 'other_class_id'")

        logger.debug(f"{self.evaluate_with_other_class} and {self.ids2labels}")
        if self.evaluate_with_other_class and self.ids2labels is None:
            # check if the selected evaluation type is valid
            logging.warning(
                "Labels to label ids correspondence is needed to make other_class specific evaluation. Since it is "
                "absent now, the standard sklearn metrics will be calculated instead."
            )
            self.evaluate_with_other_class = False

        self.max_rules = max_rules
        self.min_coverage = min_coverage
        self.drop_rules = drop_rules



import logging

from collections import Counter
from typing import Dict, List

import numpy as np

from knodle.transformation.labels import label_ids_to_labels

logger = logging.getLogger(__name__)


def classification_report_other_class(
        y_true: np.array, y_pred: np.array, ids2labels: Dict, other_class_id: int, verbose: bool = True
) -> Dict:
    """ Prepare the batch for the label-based evaluation. """
    string_prediction, string_gold = label_ids_to_labels(
        predictions=y_pred, labels=y_true, ids2labels=ids2labels
    )
    clf_report = score(
        key=string_gold, prediction=string_prediction, verbose=verbose, other_class_label=ids2labels[other_class_id]
    )
    return clf_report


def score(
        key: List[str], prediction: List[str], verbose: bool, other_class_label: str
) -> Dict[str, float]:
    """
    Computes the precision, recall and f1-score with respect to the other_class label.
    Other class is treated as a separate negative class not included into the evaluation,
    i.e. a correctly predicted "other_class" label counts as a TN rather than a TP.
    Args:
        key: List with gold string labels for the batch
        prediction: List with predicted labels for the batch
        verbose: Whether to output per-relation statistics
        other_class_label: Label of the class that should be treated as negative.
    Returns: Decision per sample. Shape: (instances, )
    """
    correct_by_relation = Counter()
    guessed_by_relation = Counter()
    gold_by_relation = Counter()
    # Loop over the data to compute a score
    for row in range(len(key)):
        gold = key[row]
        guess = prediction[row]
        if gold == other_class_label and guess == other_class_label:
            pass
        elif gold == other_class_label and guess != other_class_label:
            guessed_by_relation[guess] += 1
        elif gold != other_class_label and guess == other_class_label:
            gold_by_relation[gold] += 1
        elif gold != other_class_label and guess != other_class_label:
            guessed_by_relation[guess] += 1
            gold_by_relation[gold] += 1
            if gold == guess:
                correct_by_relation[guess] += 1

    # Print verbose information
    if verbose:
        logger.info("Per-relation statistics:")
        relations = gold_by_relation.keys()
        longest_relation = 0
        for relation in sorted(relations):
            longest_relation = max(len(relation), longest_relation)
        for relation in sorted(relations):
            # (compute the score)
            correct = correct_by_relation[relation]
            guessed = guessed_by_relation[relation]
            gold = gold_by_relation[relation]
            prec = 1.0
            if guessed > 0:
                prec = float(correct) / float(guessed)
            recall = 0.0
            if gold > 0:
                recall = float(correct) / float(gold)
            f1 = 0.0
            if prec + recall > 0:
                f1 = 2.0 * prec * recall / (prec + recall)
            # (print the score)
            line = f"{relation:{longest_relation}s}\tP: {prec:6.2f}    R: {recall:6.2f}    " \
                f"F1: {f1:6.2f}    #: {gold:6d}"
            logger.info(line)

    prec_micro = 1.0
    if sum(guessed_by_relation.values()) > 0:
        prec_micro = float(sum(correct_by_relation.values())) / float(sum(guessed_by_relation.values()))
    recall_micro = 0.0
    if sum(gold_by_relation.values()) > 0:
        recall_micro = float(sum(correct_by_relation.values())) / float(sum(gold_by_relation.values()))
    f1_micro = 0.0
    if prec_micro + recall_micro > 0.0:
        f1_micro = 2.0 * prec_micro * recall_micro / (prec_micro + recall_micro)

    # Print the aggregate score
    logger.info("Final Score:")
    logger.info("Precision (micro): {:.3%}".format(prec_micro))
    logger.info("   Recall (micro): {:.3%}".format(recall_micro))
    logger.info("       F1 (micro): {:.3%}".format(f1_micro))
    logger.info("\n")

    return {"precision": prec_micro, "recall": recall_micro, "f1": f1_micro}



import numpy as np
from numpy.testing import assert_array_equal

from torch import Tensor, equal
from torch.utils.data import TensorDataset

from knodle.transformation.torch_input import input_labels_to_tensordataset


def test_input_labels_to_tensordataset():

    a = np.ones((4, 4))
    b = np.ones((4, 3))
    labels = np.ones((4,))

    input_data = TensorDataset(Tensor(a), Tensor(b))

    input_label_dataset = input_labels_to_tensordataset(input_data, labels)

    assert len(input_label_dataset.tensors) == 3
    assert_array_equal(input_label_dataset.tensors[-1].cpu().detach().numpy(), labels)


import logging

import numpy as np
from cleanlab.classification import LearningWithNoisyLabels
from skorch import NeuralNetClassifier
from torch.utils.data import TensorDataset

from knodle.trainer import MajorityVoteTrainer
from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.cleanlab.config import CleanLabConfig
from knodle.trainer.cleanlab.latent_estimation import estimate_cv_predicted_probabilities_split_by_rules, \
    estimate_cv_predicted_probabilities_split_by_signatures
from knodle.transformation.majority import input_to_majority_vote_input
from knodle.transformation.torch_input import dataset_to_numpy_input

logger = logging.getLogger(__name__)


@AutoTrainer.register('cleanlab')
class CleanLabTrainer(MajorityVoteTrainer):
    def __init__(self, **kwargs):
        if kwargs.get("trainer_config", None) is None:
            kwargs["trainer_config"] = CleanLabConfig()
        super().__init__(**kwargs)

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ) -> None:

        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)
        self._apply_rule_reduction()

        if dev_model_input_x is not None and dev_gold_labels_y is not None:
            logger.info("Validation data is not used during Cleanlab training")

        # since CL accepts only sklearn.classifier compliant class, we wraps the PyTorch model
        self.model = NeuralNetClassifier(
            self.model,
            criterion=self.trainer_config.criterion,
            optimizer=self.trainer_config.optimizer,
            lr=self.trainer_config.lr,
            max_epochs=self.trainer_config.epochs,
            batch_size=self.trainer_config.batch_size,
            train_split=None,
            callbacks="disable",
            device=self.trainer_config.device
        )

        # calculate labels based on t and z; perform additional filtering if applicable
        self.model_input_x, noisy_y_train, self.rule_matches_z = input_to_majority_vote_input(
            self.rule_matches_z, self.mapping_rules_labels_t, self.model_input_x,
            use_probabilistic_labels=self.trainer_config.use_probabilistic_labels,
            filter_non_labelled=self.trainer_config.filter_non_labelled,
            other_class_id=self.trainer_config.other_class_id,)

        # turn input to the CL-compatible format
        model_input_x_numpy = dataset_to_numpy_input(self.model_input_x)

        # calculate psx in advance with splitting by rules
        if self.trainer_config.psx_calculation_method == "rules":
            psx = estimate_cv_predicted_probabilities_split_by_rules(
                self.model_input_x, noisy_y_train, self.rule_matches_z, self.model, self.trainer_config.output_classes,
                seed=self.trainer_config.seed, cv_n_folds=self.trainer_config.cv_n_folds
            )

        elif self.trainer_config.psx_calculation_method == "signatures":
            psx = estimate_cv_predicted_probabilities_split_by_signatures(
                self.model_input_x, noisy_y_train, self.rule_matches_z, self.model, self.trainer_config.output_classes,
                seed=self.trainer_config.seed, cv_n_folds=self.trainer_config.cv_n_folds
            )

        elif self.trainer_config.psx_calculation_method == "random":
            # if no psx calculation method is specified, psx will be calculated in CL with random folder splitting
            psx = None

        else:
            raise ValueError("Unknown psx calculation method.")

        # CL denoising and training
        rp = LearningWithNoisyLabels(
            clf=self.model, seed=self.trainer_config.seed,
            cv_n_folds=self.trainer_config.cv_n_folds,
            prune_method=self.trainer_config.prune_method,
            converge_latent_estimates=self.trainer_config.converge_latent_estimates,
            pulearning=self.trainer_config.pulearning,
            n_jobs=self.trainer_config.n_jobs
        )
        _ = rp.fit(model_input_x_numpy, noisy_y_train, psx=psx)
        logging.info("Training is done.")




"""knodle - Knowledge infused deep learning framework"""

__author__ = "knodle <knodle@cs.univie.ac.at>"
__all__ = []
from knodle.version import __version__

import logging

logger = logging.getLogger()
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)

logger.info("Initalized logger")



import logging
import os
from copy import copy

import numpy as np
import torch
from joblib import load
from torch.nn import Module
from torch.optim import SGD
from torch.utils.data import TensorDataset

from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.baseline.majority import MajorityVoteTrainer
from knodle.trainer.wscrossweigh.config import WSCrossWeighConfig
from knodle.trainer.wscrossweigh.wscrossweigh_weights_calculator import WSCrossWeighWeightsCalculator

from knodle.transformation.filter import filter_empty_probabilities
from knodle.transformation.majority import z_t_matrices_to_majority_vote_probs
from knodle.transformation.torch_input import input_info_labels_to_tensordataset

torch.set_printoptions(edgeitems=100)
logger = logging.getLogger(__name__)
logging.getLogger('matplotlib.font_manager').disabled = True


@AutoTrainer.register('wscrossweigh')
class WSCrossWeighTrainer(MajorityVoteTrainer):

    def __init__(
            self,
            cw_model: Module = None,
            cw_model_input_x: TensorDataset = None,
            cw_rule_matches_z: np.ndarray = None,
            run_classifier: bool = True,  # set to False if you want only the calculation of the sample weights
            use_weights: bool = True,  # set to False if you want to use weights = 1 (baseline)
            **kwargs
    ):
        self.cw_model = cw_model if cw_model else kwargs.get("model")
        self.cw_model_input_x = cw_model_input_x if cw_model_input_x else kwargs.get("model_input_x")
        self.cw_rule_matches_z = cw_rule_matches_z if cw_rule_matches_z else kwargs.get("rule_matches_z")

        if kwargs.get("trainer_config") is None:
            kwargs["trainer_config"] = WSCrossWeighConfig(
                optimizer=SGD,
                cw_optimizer=SGD,
                lr=0.001,
                cw_lr=0.001,
            )
        super().__init__(**kwargs)

        self.run_classifier = run_classifier
        self.use_weights = use_weights

        logger.info("CrossWeigh Config is used: {}".format(self.trainer_config.__dict__))

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        """ This function sample_weights the samples with WSCrossWeigh method and train the model """
        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)
        self._apply_rule_reduction()

        # initialise optimizer
        self.trainer_config.optimizer = self.initialise_optimizer()

        train_labels = self.calculate_labels()

        sample_weights = self._get_sample_weights() if self.use_weights \
            else torch.FloatTensor([1] * len(self.model_input_x))

        if not self.run_classifier:
            logger.info("No classifier is to be trained")
            return
        logger.info("Classifier training is started")

        train_loader = self._make_dataloader(
            input_info_labels_to_tensordataset(self.model_input_x, sample_weights.cpu().detach().numpy(), train_labels)
        )

        self._train_loop(train_loader, use_sample_weights=True, draw_plot=self.trainer_config.draw_plot)

    def calculate_labels(self) -> np.ndarray:
        """ This function calculates label probabilities and filter out non labelled samples, when needed """
        train_labels = z_t_matrices_to_majority_vote_probs(
            self.rule_matches_z, self.mapping_rules_labels_t, self.trainer_config.other_class_id
        )

        if self.trainer_config.filter_non_labelled:
            self.model_input_x, train_labels, self.rule_matches_z = filter_empty_probabilities(
                self.model_input_x, train_labels, self.rule_matches_z
            )

        if train_labels.shape[1] != self.trainer_config.output_classes:
            raise ValueError(
                f"The number of output classes {self.trainer_config.output_classes} do not correspond to labels "
                f"probabilities dimension {train_labels.shape[1]}"
            )

        return train_labels

    def _get_sample_weights(self) -> torch.FloatTensor:
        """ This function checks whether there are accessible already pretrained sample weights. If yes, return
        them. If not, calculates sample weights calling method of WSCrossWeighWeightsCalculator class"""

        if os.path.isfile(os.path.join(
                self.trainer_config.caching_folder, f"sample_weights_{self.trainer_config.caching_suffix}.lib")
        ):
            logger.info("Already pretrained samples sample_weights will be used.")
            sample_weights = load(os.path.join(
                self.trainer_config.caching_folder, f"sample_weights_{self.trainer_config.caching_suffix}.lib")
            )
        else:
            logger.info("No pretrained sample weights are found, they will be calculated now")
            sample_weights = WSCrossWeighWeightsCalculator(
                model=self.cw_model,
                mapping_rules_labels_t=self.mapping_rules_labels_t,
                model_input_x=self.cw_model_input_x,
                rule_matches_z=self.cw_rule_matches_z,
                trainer_config=self.get_denoising_config(),
            ).calculate_weights()
            logger.info(f"Sample weights are calculated and saved to {self.trainer_config.caching_folder} folder")
        return sample_weights

    def get_denoising_config(self):
        """ Get a config for WSCrossWeigh sample weights calculation """
        weights_calculation_config = copy(self.trainer_config)
        weights_calculation_config.epochs = self.trainer_config.cw_epochs
        weights_calculation_config.optimizer = self.trainer_config.cw_optimizer
        weights_calculation_config.lr = self.trainer_config.cw_lr
        weights_calculation_config.batch_size = self.trainer_config.cw_batch_size
        weights_calculation_config.filter_non_labelled = self.trainer_config.cw_filter_non_labelled
        weights_calculation_config.other_class_id = self.trainer_config.cw_other_class_id
        weights_calculation_config.grad_clipping = self.trainer_config.cw_grad_clipping
        weights_calculation_config.seed = self.trainer_config.cw_seed
        weights_calculation_config.saved_models_dir = None
        return weights_calculation_config



from knodle.trainer.auto_config import AutoConfig
from knodle.trainer.config import BaseTrainerConfig


@AutoConfig.register("majority")
class MajorityConfig(BaseTrainerConfig):
    def __init__(
            self,
            use_probabilistic_labels: bool = True,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.use_probabilistic_labels = use_probabilistic_labels
        self.probability_threshold = None



import logging
import random
from typing import List, Dict, Union, Tuple

import scipy.sparse as sp
import numpy as np
import torch
from torch.utils.data import TensorDataset

from knodle.trainer.wscrossweigh.utils import return_unique
from knodle.transformation.torch_input import input_info_labels_to_tensordataset, input_labels_to_tensordataset

logger = logging.getLogger(__name__)


def k_folds_splitting_by_rules(
        data_features: TensorDataset, labels: np.ndarray, rule_matches_z: np.ndarray, partitions: int, num_folds: int,
        seed: int = None, other_class_id: int = None
) -> Tuple[List, List]:
    """
    This function allows to perform the splitting of data instances into k folds according to the rules matched
    in them. The splitting could be performed in several iterations (defined by "partition" parameter).
    The logic is the following:
        for each partition:
        - the rules are shuffled
        - the rules are splitted into k folds
        - each fold iteratively becomes a hold-out fold
        - the samples that matched the rules from the hold-out fold are added to the hold-out test set
        - other samples are added to the train set
    The train and test sets do not intersect. If a sample matched several rules and one of the rules is included
    in the test set (and therefore the sample as well), this sample won't be included in the training set.
    Correspondingly, such samples (where multiple rules matched) are included into several hold-out folds depending on
    which of the relevant rules is selected to the hold-out fold in the current splitting.

    :param data_features: encoded data samples
    :param labels: array of labels (num_samples x 1)
    :param rule_matches_z: matrix of rule matches (num_samples x num_rules)
    :param partitions: number of partitions that are to be performed; in each partition the dataset will be splitted
    into k folds
    :param num_folds: number of folds the data instances are to be slitted into in each partition
    :param seed: optionally, the seed could be fixed in order to provide reproducibility
    :param other_class_id: if you don't want to include the negative samples (the ones that belong to the other class)
     to the test set, but only to the training set, you can pass the id of other class
    should be

    :return: two lists; the first contains the training sets, the second contains the test (hold-out) sets.
    """

    random.seed(seed) if seed is not None else random.choice(range(9999))
    rule_id2samples_ids = get_rules_sample_ids(rule_matches_z)

    return compose_train_n_test_datasets(
        data_features, rule_id2samples_ids, labels, num_folds, partitions, other_class_id
    )


def k_folds_splitting_by_signatures(
        data_features: np.ndarray, labels: np.ndarray, rule_matches_z: np.ndarray, partitions: int, num_folds: int,
        seed: int = None, other_class_id: int = None
) -> Tuple[List, List]:
    """
    This function allows to perform the splitting of data instances into k folds according to the signatures.
    The sample signature is composed from the rule matched in the sample. For example, if rules with ids 1, 5, 7 matched
    in the sample, the sample signature is 1_5_7. Thus, the signature serves as sample identifier.
    The logic is the following:
        - the sample signatures are calculated
        for each partition:
        - the sample signatures are shuffled
        - the sample signatures are splitted into k folds
        - each fold iteratively becomes a hold-out fold
        - the samples with signatures from the hold-out fold are added to the hold-out test set
        - other samples are added to the train set
    The train and test sets do not intersect.

    :param data_features: encoded data samples (num_samples x num_features)
    :param labels: array of labels (num_samples x 1)
    :param rule_matches_z: matrix of rule matches (num_samples x num_rules)
    :param partitions: number of partitions that are to be performed; in each partition the dataset will be splitted
    into k folds
    :param num_folds: number of folds the data instances are to be slitted into in each partition
    :param seed: optionally, the seed could be fixed in order to provide reproducibility
    :param other_class_id: if you don't want to include the negative samples (the ones that belong to the other class)
     to the test set, but only to the training set, you can pass the id of other class should be

    :return: two lists; the first contains the training sets, the second contains the test (hold-out) sets.
    """

    random.seed(seed) if seed is not None else random.choice(range(9999))
    signature2samples = get_signature_sample_ids(rule_matches_z)

    return compose_train_n_test_datasets(
        data_features, signature2samples, labels, num_folds, partitions, other_class_id
    )


def get_rules_sample_ids(rule_matches_z: Union[np.ndarray, sp.csr_matrix]) -> Dict[str, List[int]]:
    """
    This function creates a dictionary {rule id : sample id where this rule matched}. The dictionary is needed as a
    support tool for faster calculation of train and test sets.
    :param rule_matches_z: matrix of rule matches as a numpy or sparse matrix (num_samples x num_rules)
    :return: dictionary {rule_id: [sample_id1, sample_id2, ...]}
    """
    if isinstance(rule_matches_z, sp.csr_matrix):
        rule2sample_id = {key: set() for key in range(rule_matches_z.shape[1])}
        for row_id, col_id in zip(*rule_matches_z.nonzero()):       # row_id = sample id, col_id = rule id
            rule2sample_id[col_id].add(row_id)
    else:
        rule2sample_id = dict((i, set()) for i in range(0, rule_matches_z.shape[1]))
        for row_id, row in enumerate(rule_matches_z):
            rule_ids = np.where(row == 1)[0].tolist()
            for rule_id in rule_ids:
                rule2sample_id[rule_id].add(row_id)
    return rule2sample_id


def get_signature_sample_ids(rule_matches_z: np.ndarray) -> Dict:
    """
    This function calculates the signature for each sample (e.g. if rules with ids 1, 5, 7 matched in a sample, a sample
    signature id is 1_5_7) and id for each signature.
    :param rule_matches_z: matrix of rule matches (num_samples x num_rules)
    :return: two dictionaries:
        - {signature: signature id}
        - {signature id: [sample_id1, sample_id2, ...]}
    """

    signature2samples = {}

    if isinstance(rule_matches_z, sp.csr_matrix):
        samples_id_rules_dict = {key: [] for key in range(rule_matches_z.shape[0])}
        for row_id, col_id in zip(*rule_matches_z.nonzero()):
            samples_id_rules_dict[row_id].append(col_id)

    else:
        samples_id_rules_dict = dict((i, set()) for i in range(0, rule_matches_z.shape[0]))
        for row_id, row in enumerate(rule_matches_z):
            samples_id_rules_dict[row_id] = np.where(row == 1)[0].tolist()

    for sample_id, rules in samples_id_rules_dict.items():
        signature = "_".join(map(str, sorted(list(rules))))
        signature2samples.setdefault(signature, []).append(sample_id)

    return signature2samples


def compose_train_n_test_datasets(
        data_features: TensorDataset, rule2samples: Dict, labels: np.ndarray, num_folds: int,
        partitions: int, other_class_id: int = None
) -> Tuple[List, List]:
    """
    This function creates train and test datasets for k-folds cross-validation.

    :param data_features: encoded data samples
    :param rule2samples: {rule: [sample_id1, sample_id2, ...]}. Rule in this context means anything basing on
    which splitting is performed (matched rule, sample signature, ...).
    :param labels: array of labels
    :param num_folds: number of folds the data instances are to be slitted into in each partition
    :param partitions: number of partitions that are to be performed
    :param other_class_id: if you don't want to include the negative samples (the ones that belong to the other class)
     to the test set, but only to the training set, you can pass the id of other class should be

    :return: list of train sets and list of corresponding test (hold-out) sets
    """

    # calculate ids of all samples that belong to the 'other_class'
    other_sample_ids = np.where(labels[:, other_class_id] == 1)[0].tolist() if other_class_id else None
    # make a list of rule ids to shuffle them later
    rule_ids = list(rule2samples.keys())

    train_datasets, test_datasets = [], []
    for partition in range(partitions):
        logger.info(f"Partition {partition + 1}/{partitions}:")
        random.shuffle(rule_ids)  # shuffle anew for each splitting
        for fold_id in range(num_folds):
            train_dataset, test_dataset = get_train_test_datasets_by_rule_indices(
                data_features, rule_ids, rule2samples, labels, fold_id, num_folds, other_sample_ids
            )
            train_datasets.append(train_dataset)
            test_datasets.append(test_dataset)

    return train_datasets, test_datasets


def get_train_test_datasets_by_rule_indices(
        data_features: TensorDataset, rules_ids: List[int], rule2samples: Dict, labels: np.ndarray, fold_id: int,
        num_folds: int, other_sample_ids: List[int]
) -> Tuple[TensorDataset, TensorDataset]:
    """
    This function returns train and test datasets for k-fold cross validation training. Each dataloader comprises
    encoded samples, labels and sample indices in the original matrices.

    :param data_features: encoded data samples
    :param rules_ids: list of shuffled rules indices
    :param rule2samples: dictionary that contains information about corresponding between rules ids and sample ids.
    Rule in this context means anything basing on which splitting is performed (matched rule, sample signature, ...)
    :param labels: labels of all training samples (num_samples, num_classes)
    :param fold_id: number of a current hold-out fold
    :param num_folds: the whole number of folds the data should be splitted into
    :param other_sample_ids: a list of sample ids that belong to the other class. They won't be included in the test
    set, but only to the training set.
    should be

    :return: dataloaders for cw training and testing
    """
    train_rules, test_rules = calculate_rules_indices(rules_ids, fold_id, num_folds)

    # select train and test samples and labels according to the selected rules idx
    test_samples, test_labels, test_idx = get_samples_labels_idx_by_rule_id(
        data_features, labels, test_rules, rule2samples, check_intersections=None,
        other_sample_ids=other_sample_ids
    )

    train_samples, train_labels, _ = get_samples_labels_idx_by_rule_id(
        data_features, labels, train_rules, rule2samples, check_intersections=test_idx,
        other_sample_ids=other_sample_ids
    )

    train_dataset = input_labels_to_tensordataset(train_samples, train_labels)
    test_dataset = input_info_labels_to_tensordataset(test_samples, test_idx, test_labels)

    logger.info(
        f"Fold {fold_id}     Rules in training set: {len(train_rules)}, rules in test set: {len(test_rules)}, "
        f"samples in training set: {len(train_samples)}, samples in test set: {len(test_samples)}"
    )

    return train_dataset, test_dataset


def calculate_rules_indices(rules_idx: list, fold_id: int, num_folds: int) -> Tuple[list, list]:
    """
    Calculates the indices of the samples which are to be included in training and test sets for k-fold cross validation

    :param rules_idx: all rules indices (shuffled) that are to be splitted into cw training & cw test set rules
    :param fold_id: number of a current hold-out fold
    :param num_folds: the whole number of folds the data should be splitted into

    :return: two arrays containing indices of rules that will be used for cw training and cw test set accordingly
    """
    test_rules_idx = rules_idx[fold_id::num_folds]
    train_rules_idx = [rule_id for rule_id in rules_idx if rule_id not in test_rules_idx]

    if not set(test_rules_idx).isdisjoint(set(train_rules_idx)):
        raise ValueError("Splitting into train and test rules is done incorrectly.")

    return train_rules_idx, test_rules_idx


def get_samples_labels_idx_by_rule_id(
        data_features: TensorDataset, labels: np.ndarray, indices: list, rule2samples: Dict,
        check_intersections: np.ndarray = None, other_sample_ids: list = None
) -> Tuple[TensorDataset, np.ndarray, np.ndarray]:
    """
    Extracts the samples and labels from the original matrices by indices. If intersection is filled with
    another sample matrix, it also checks whether the sample is not in this other matrix yet.

    :param data_features: encoded data samples
    :param labels: all training samples labels (num_samples, num_classes)
    :param indices: indices of rules; samples, where these rules matched & their labels are to be included in set
    :param rule2samples: dictionary that contains information about corresponding from rules to sample ids.
    :param check_intersections: optional parameter that indicates that intersections should be checked (used to
    exclude the sentences from the training set which are already in the test set)
    :param other_sample_ids: a list of sample ids that belong to the other class. They won't be included in the test
    set, but only to the training set.

    :return: samples, labels and indices in the original matrix
    """
    sample_ids = [list(rule2samples.get(idx)) for idx in indices]
    sample_ids = list(set([value for sublist in sample_ids for value in sublist]))

    if other_sample_ids is not None:
        sample_ids = list(set(sample_ids).union(set(other_sample_ids)))

    if check_intersections is not None:
        sample_ids = return_unique(np.array(sample_ids), check_intersections)

    # samples_dataset = TensorDataset(torch.Tensor(data_features.tensors[0][sample_ids]))
    samples_dataset = TensorDataset(*[inp[sample_ids] for inp in data_features.tensors])
    samples_labels = np.array(labels[sample_ids])
    samples_idx = np.array(sample_ids)

    return samples_dataset, samples_labels, samples_idx



import argparse
import os
import sys

import joblib
import pandas as pd
from minio import Minio
from torch import Tensor, LongTensor
from torch.optim import Adam
from torch.utils.data import TensorDataset
from tqdm import tqdm
from transformers import DistilBertTokenizer, DistilBertForSequenceClassification, AdamW
from examples.trainer.preprocessing import convert_text_to_transformer_input, get_tfidf_features
from knodle.model.logistic_regression_model import LogisticRegressionModel
from knodle.trainer.wscrossweigh.config import WSCrossWeighConfig
from knodle.trainer.wscrossweigh.wscrossweigh import WSCrossWeighTrainer


def train_wscrossweigh(path_to_data: str, num_classes: int) -> None:
    """
    We are going to train a BERT classification model using weakly annotated data with additional WSCrossWeigh
    denoising. The sample weights in WSCrossWeigh will be trained with logistic regression in order to, firstly,
    reduce the computational effort, and, secondly, demonstrate the ability of the algorithm to use different models
    for data denoising and classifier training.
    :param path_to_data: path to the folder where all the input data is stored
    :param num_classes: number of output classes
    """

    num_classes = int(num_classes)

    # Define constants
    imdb_data_dir = os.path.join(os.getcwd(), "datasets", "imdb")
    processed_data_dir = os.path.join(imdb_data_dir, "processed")
    os.makedirs(processed_data_dir, exist_ok=True)

    # Download data
    client = Minio("knodle.dm.univie.ac.at", secure=False)
    files = [
        "df_train.csv", "df_dev.csv", "df_test.csv",
        "train_rule_matches_z.lib", "dev_rule_matches_z.lib", "test_rule_matches_z.lib",
        "mapping_rules_labels_t.lib"
    ]
    for file in tqdm(files):
        client.fget_object(
            bucket_name="knodle",
            object_name=os.path.join("datasets/imdb/processed", file),
            file_path=os.path.join(processed_data_dir, file),
        )

    # Load data into memory
    df_train = pd.read_csv(os.path.join(processed_data_dir, "df_train.csv"))
    df_dev = pd.read_csv(os.path.join(processed_data_dir, "df_dev.csv"))
    df_test = pd.read_csv(os.path.join(processed_data_dir, "df_test.csv"))

    mapping_rules_labels_t = joblib.load(os.path.join(processed_data_dir, "mapping_rules_labels_t.lib"))

    train_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "train_rule_matches_z.lib"))

    # sample weights are calculated with logistic regression model (with TF-IDF features); the BERT model is used for
    # the final classifier training.
    train_tfidf_sparse, dev_tfidf_sparse, _ = get_tfidf_features(df_train["sample"].tolist(), df_dev["sample"].tolist())
    train_tfidf = Tensor(train_tfidf_sparse.toarray())
    train_dataset_tfidf = TensorDataset(train_tfidf)

    # For the BERT training we convert train, dev, test data to BERT encoded features (input indices & attention mask)
    model_name = 'distilbert-base-uncased'
    tokenizer = DistilBertTokenizer.from_pretrained(model_name)

    X_train = convert_text_to_transformer_input(df_train["sample"].tolist(), tokenizer)
    X_dev = convert_text_to_transformer_input(df_dev["sample"].tolist(), tokenizer)
    X_test = convert_text_to_transformer_input(df_test["sample"].tolist(), tokenizer)

    y_dev = TensorDataset(LongTensor(df_dev["label"].tolist()))
    y_test = TensorDataset(LongTensor(df_test["label"].tolist()))

    # define the all needed parameters in a dictionary for convenience (can also be directly passed to Trainer/Config)
    parameters = {
        "lr": 1e-4, "cw_lr": 0.8, "epochs": 5, "cw_partitions": 2, "cw_folds": 5, "cw_epochs": 2, "weight_rr": 0.7,
        "samples_start_weights": 4.0
    }
    # to have sample weights saved with some specific index in the file name, you can use "caching_suffix" variable
    caching_suffix = f"dscw_{parameters.get('cw_partitions')}part_{parameters.get('cw_folds')}folds_" \
                     f"{parameters.get('weight_rr')}wrr"

    # define LogReg and BERT models for training sample weights and final classifier correspondingly
    cw_model = LogisticRegressionModel(train_tfidf.shape[1], num_classes)
    model = DistilBertForSequenceClassification.from_pretrained('distilbert-base-uncased', num_labels=num_classes)

    # define a custom WSCrossWeigh config. If no custom config is defined, the WSCrossWeighTrainer will use the default
    # WSCrossWeighConfig which is stored in the fold with the WSCrossWeigh trainer
    custom_wscrossweigh_config = WSCrossWeighConfig(
        # general trainer parameters
        output_classes=num_classes,
        filter_non_labelled=False,
        other_class_id=3,
        seed=12345,
        epochs=parameters.get("epochs"),
        batch_size=16,
        optimizer=AdamW,
        lr=parameters.get("lr"),
        grad_clipping=5,
        caching_suffix=caching_suffix,
        saved_models_dir=os.path.join(path_to_data, "trained_models"),  # trained classifier model will be saved after each epoch

        # WSCrossWeigh specific parameters
        partitions=parameters.get("cw_partitions"),  # number of WSCrossWeigh iterations (= splitting into folds)
        folds=parameters.get("cw_folds"),  # number of folds train data will be splitted into
        weight_reducing_rate=parameters.get("weight_rr"),  # sample weights reducing coefficient
        samples_start_weights=parameters.get("samples_start_weights"),  # the start weight of sample weights
        cw_epochs=parameters.get("cw_epochs"),  # number of epochs each WSCrossWeigh model is to be trained
        cw_optimizer=Adam,  # WSCrossWeigh model optimiser
        cw_lr=parameters.get("cw_lr")  # WSCrossWeigh model lr
    )

    trainer = WSCrossWeighTrainer(
        # general Trainer inputs (a more detailed explanation of Knodle inputs is in README)
        model=model,  # classification model
        mapping_rules_labels_t=mapping_rules_labels_t,  # t matrix
        model_input_x=X_train,  # x matrix for training the classifier
        rule_matches_z=train_rule_matches_z,  # z matrix
        trainer_config=custom_wscrossweigh_config,

        # additional dev set used for classification model evaluation during training
        dev_model_input_x=X_dev,
        dev_gold_labels_y=y_dev,

        # WSCrossWeigh specific parameters. If they are not defined, the corresponding main classification parameters
        # will be used instead (model instead of cw_model etc)
        cw_model=cw_model,  # model that will be used for WSCrossWeigh weights calculation
        cw_model_input_x=train_dataset_tfidf,  # x matrix for training the WSCrossWeigh models
    )

    # the WSCrossWeighTrainer is trained
    trainer.train()
    # the trained model is tested on the test set
    clf_report, _ = trainer.test(X_test, y_test)
    print(clf_report)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))
    parser.add_argument("--path_to_data", help="Path to the folder where all input files are stored.")
    parser.add_argument("--num_classes", help="Number of classes")
    args = parser.parse_args()

    train_wscrossweigh(args.path_to_data, args.num_classes)



import os
from typing import Union, List, Tuple

import pandas as pd
import numpy as np
from joblib import load


def read_train_dev_test(
        target_path: str, if_dev_data: bool = False
) -> Union[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray, np.ndarray],
           Tuple[pd.DataFrame, None, pd.DataFrame, np.ndarray, np.ndarray, np.ndarray]]:
    """This function loads the matrices as well as train, test (occasionally, also dev) data from corresponding files"""
    df_train = load(os.path.join(target_path, 'df_train.lib'))
    df_test = load(os.path.join(target_path, 'df_test.lib'))
    z_train_rule_matches = load(os.path.join(target_path, 'train_rule_matches_z.lib'))
    z_test_rule_matches = load(os.path.join(target_path, 'test_rule_matches_z.lib'))
    t_mapping_rules_labels = load(os.path.join(target_path, 'mapping_rules_labels_t.lib'))

    if if_dev_data:
        dev_df = load(os.path.join(target_path, 'df_dev.lib'))
        return df_train, dev_df, df_test, z_train_rule_matches, z_test_rule_matches, t_mapping_rules_labels

    return df_train, None, df_test, z_train_rule_matches, z_test_rule_matches, t_mapping_rules_labels


def get_samples_list(data: Union[pd.Series, pd.DataFrame], column_num: int = None) -> List:
    """ Extracts the data from the Series/DataFrame and returns it as a list"""
    column_num = int(column_num)
    if isinstance(data, pd.Series):
        return list(data)
    elif isinstance(data, pd.DataFrame) and column_num is not None:
        return list(data.iloc[:, column_num])
    else:
        raise ValueError(
            "Please pass input data either as a Series or as a DataFrame with number of the column with samples"
        )



import numpy as np
from scipy.sparse import csr_matrix

from knodle.transformation.rule_reduction import _get_merged_matrix, reduce_rule_matches, _get_rule_by_label_iterator


def test_reduction():
    # test rule iterator
    mapping_rule_class_t = np.array([
        [1, 0],
        [0, 1],
        [1, 0],
        [1, 0]])
    rule_iterator = list(_get_rule_by_label_iterator(mapping_rule_class_t))
    expected_iterator = [
        np.array([0, 2, 3]),
        np.array([1])
    ]

    assert len(rule_iterator) == len(expected_iterator)
    assert np.array_equal(rule_iterator[0], expected_iterator[0])
    assert np.array_equal(rule_iterator[1], expected_iterator[1])

    # test _get_merged_matrix
    rule_matches_z = np.array([
        [0, 1, 0, 1],
        [1, 0, 1, 1],
        [0, 0, 1, 0],
        [1, 0, 1, 0],
        [0, 0, 1, 0]])
    to_keep_mask = np.array([1, 2])

    merged_rule_matches_z = _get_merged_matrix(
        full_matches=rule_matches_z, to_keep_mask=to_keep_mask, label_rule_masks=expected_iterator)
    expected_merged = np.array([
        [1],
        [1],
        [0],
        [1],
        [0]])
    assert np.array_equal(merged_rule_matches_z, expected_merged)

    # test end-to-end merge
    test_rule_matches_z = np.array([
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [1, 0, 0, 0]])

    out = reduce_rule_matches(
        rule_matches_z=rule_matches_z, mapping_rules_labels_t=mapping_rule_class_t,
        rule_matches_rest={"test_matches": test_rule_matches_z},
        drop_rules=False, max_rules=2, min_coverage=1.0)

    expected = {
        "train_rule_matches_z": np.array([
            [1, 1],
            [1, 0],
            [1, 0],
            [1, 0],
            [1, 0]
        ]),
        "test_matches": np.array([
            [0, 1],
            [1, 0],
            [1, 0]
        ]),
        "mapping_rules_labels_t": np.array([
            [1, 0],
            [0, 1]
        ])
    }

    assert np.array_equal(out.get("train_rule_matches_z"), expected["train_rule_matches_z"])
    assert np.array_equal(out.get("test_matches"), expected["test_matches"])
    assert np.array_equal(out.get("mapping_rules_labels_t"), expected["mapping_rules_labels_t"])

    # test end-to-end by drop
    out = reduce_rule_matches(
        rule_matches_z=rule_matches_z, mapping_rules_labels_t=mapping_rule_class_t,
        rule_matches_rest={"test_matches": test_rule_matches_z},
        drop_rules=True, max_rules=2, min_coverage=0.0)

    expected = {
        "train_rule_matches_z": np.array([
            [0, 0],
            [1, 1],
            [0, 1],
            [1, 1],
            [0, 1]
        ]),
        "test_matches": np.array([
            [0, 0],
            [0, 1],
            [1, 0]
        ]),
        "mapping_rules_labels_t": np.array([
            [1, 0],
            [1, 0]
        ])
    }

    assert np.array_equal(out.get("train_rule_matches_z"), expected["train_rule_matches_z"])
    assert np.array_equal(out.get("test_matches"), expected["test_matches"])
    assert np.array_equal(out.get("mapping_rules_labels_t"), expected["mapping_rules_labels_t"])


def test_reduction_for_sparse():
    # test rule iterator
    mapping_rule_class_t = csr_matrix([
        [1, 0],
        [0, 1],
        [1, 0],
        [1, 0]])
    rule_iterator = list(_get_rule_by_label_iterator(mapping_rule_class_t))
    expected_iterator = [
        np.array([0, 2, 3]),
        np.array([1])
    ]

    assert len(rule_iterator) == len(expected_iterator)
    assert np.array_equal(rule_iterator[0], expected_iterator[0])
    assert np.array_equal(rule_iterator[1], expected_iterator[1])

    # test _get_merged_matrix
    rule_matches_z = csr_matrix([
        [0, 1, 0, 1],
        [1, 0, 1, 1],
        [0, 0, 1, 0],
        [1, 0, 1, 0],
        [0, 0, 1, 0]])
    to_keep_mask = np.array([1, 2])

    merged_rule_matches_z = _get_merged_matrix(
        full_matches=rule_matches_z, to_keep_mask=to_keep_mask, label_rule_masks=expected_iterator)
    expected_merged = csr_matrix([
        [1],
        [1],
        [0],
        [1],
        [0]])
    assert (merged_rule_matches_z != expected_merged).nnz == 0

    # test end-to-end merge
    test_rule_matches_z = csr_matrix([
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [1, 0, 0, 0]])

    out = reduce_rule_matches(
        rule_matches_z=rule_matches_z, mapping_rules_labels_t=mapping_rule_class_t,
        rule_matches_rest={"test_matches": test_rule_matches_z},
        drop_rules=False, max_rules=2, min_coverage=1.0)

    expected = {
        "train_rule_matches_z": csr_matrix([
            [1, 1],
            [1, 0],
            [1, 0],
            [1, 0],
            [1, 0]
        ]),
        "test_matches": csr_matrix([
            [0, 1],
            [1, 0],
            [1, 0]
        ]),
        "mapping_rules_labels_t": csr_matrix([
            [1, 0],
            [0, 1]
        ])
    }

    assert (out.get("train_rule_matches_z") != expected["train_rule_matches_z"]).nnz == 0
    assert (out.get("test_matches") != expected["test_matches"]).nnz == 0
    assert (out.get("mapping_rules_labels_t") != expected["mapping_rules_labels_t"]).nnz == 0
    assert isinstance(out.get("train_rule_matches_z"), csr_matrix)
    assert isinstance(out.get("mapping_rules_labels_t"), csr_matrix)

    # test end-to-end by drop with sparse matches and dense mapping T
    mapping_rule_class_t = np.array([
        [1, 0],
        [0, 1],
        [1, 0],
        [1, 0]])

    out = reduce_rule_matches(
        rule_matches_z=rule_matches_z, mapping_rules_labels_t=mapping_rule_class_t,
        rule_matches_rest={"test_matches": test_rule_matches_z},
        drop_rules=True, max_rules=2, min_coverage=0.0)

    expected = {
        "train_rule_matches_z": csr_matrix([
            [0, 0],
            [1, 1],
            [0, 1],
            [1, 1],
            [0, 1]
        ]),
        "test_matches": csr_matrix([
            [0, 0],
            [0, 1],
            [1, 0]
        ]),
        "mapping_rules_labels_t": np.array([
            [1, 0],
            [1, 0]
        ])
    }

    assert (out.get("train_rule_matches_z") != expected["train_rule_matches_z"]).nnz == 0
    assert (out.get("test_matches") != expected["test_matches"]).nnz == 0
    assert np.array_equal(out.get("mapping_rules_labels_t"), expected["mapping_rules_labels_t"])
    assert isinstance(out.get("train_rule_matches_z"), csr_matrix)
    assert isinstance(out.get("mapping_rules_labels_t"), np.ndarray)




from torch.optim import Optimizer

from knodle.trainer.baseline.config import MajorityConfig
from knodle.trainer.auto_config import AutoConfig


@AutoConfig.register("wscrossweigh")
class WSCrossWeighConfig(MajorityConfig):
    def __init__(
            self,
            partitions: int = 2,
            folds: int = 10,
            weight_reducing_rate: float = 0.5,
            samples_start_weights: float = 2.0,
            cw_epochs: int = None,
            cw_batch_size: int = None,
            cw_optimizer: Optimizer = None,
            cw_lr: int = 0.1,
            cw_filter_non_labelled: bool = None,
            cw_other_class_id: int = None,
            cw_grad_clipping: int = None,
            cw_seed: int = None,
            draw_plot: bool = False,
            **kwargs
    ):
        """
        A default configuration of WSCrossWeigh Trainer.

        :param partitions: number of times the sample weights calculation procedure will be performed
        (with different folds splitting)
        :param folds: number of folds the samples will be splitted into in each sample weights calculation iteration
        :param weight_reducing_rate: a value the sample weight is reduced by each time the sample is misclassified
        :param samples_start_weights: a start weight of all samples
        :param cw_epochs: number of epochs WSCrossWeigh models are to be trained
        :param cw_batch_size: batch size for WSCrossWeigh models training
        :param cw_optimizer: optimizer for WSCrossWeigh models training
        :param cw_lr: learning rate for WSCrossWeigh models training
        :param cw_filter_non_labelled: whether the samples where no rule matched are to be filtered out in WSCrossWeigh
        :param cw_other_class_id: id of the negative class; if set, the samples with no rule matched will be assigned to
        it in WSCrossWeigh
        :param cw_grad_clipping: if set to True, gradient norm of an iterable of parameters will be clipped in WSCrossWeigh
        :param cw_seed: the desired seed for generating random numbers in WSCrossWeigh
        :param draw_plot: draw a plot of development data (accuracy & loss)
        """

        super().__init__(**kwargs)
        self.draw_plot = draw_plot
        self.partitions = partitions
        self.folds = folds
        self.weight_reducing_rate = weight_reducing_rate
        self.samples_start_weights = samples_start_weights

        if cw_grad_clipping is None:
            self.cw_grad_clipping = self.grad_clipping
        else:
            self.cw_grad_clipping = cw_grad_clipping

        if cw_epochs is None:
            self.cw_epochs = self.epochs
        else:
            self.cw_epochs = cw_epochs

        if cw_batch_size is None:
            self.cw_batch_size = self.batch_size
        else:
            self.cw_batch_size = cw_batch_size

        if cw_optimizer is None:
            self.cw_optimizer = self.optimizer
        else:
            self.cw_optimizer = cw_optimizer

        if cw_filter_non_labelled is None and cw_other_class_id is None:
            self.cw_filter_non_labelled = self.filter_non_labelled
            self.cw_other_class_id = self.other_class_id
        else:
            self.cw_filter_non_labelled = cw_filter_non_labelled
            self.cw_other_class_id = cw_other_class_id

        self.cw_seed = cw_seed
        self.cw_lr = cw_lr




from tests.trainer.generic import std_trainer_input_1

from knodle.trainer.multi_trainer import MultiTrainer


def test_auto_train(std_trainer_input_1):
    (
        model,
        model_input_x, rule_matches_z, mapping_rules_labels_t,
        y_labels
    ) = std_trainer_input_1

    trainers = ["majority", "snorkel", "knn", "snorkel_knn"]
    trainer = MultiTrainer(
        name=trainers,
        model=model,
        mapping_rules_labels_t=mapping_rules_labels_t,
        model_input_x=model_input_x,
        rule_matches_z=rule_matches_z,
    )

    trainer.train()
    metrics = trainer.test(model_input_x, y_labels)

    # Check whether the code ran up to here
    assert 2 == 2



import numpy as np

from knodle.evaluation.majority import majority_sklearn_report


def test_majority_vote_no_match():
    z = np.zeros((2, 4))
    t = np.zeros((4, 2))

    z[0, 0] = 1
    z[1, 1] = 1

    t[0, 0] = 1
    t[1, 0] = 1
    t[2, 0] = 1
    t[3, 0] = 1

    y = np.array([1, 1])
    report = majority_sklearn_report(z, t, y)
    assert report["accuracy"] == 0


def test_majority_vote_correctness():
    z = np.zeros((2, 4))
    t = np.zeros((4, 2))

    z[0, 0] = 1
    z[1, 1] = 1

    t[0, 0] = 1
    t[1, 1] = 1
    t[2, 0] = 1
    t[3, 0] = 1

    y = np.array([1, 1])
    report = majority_sklearn_report(z, t, y)
    assert report["accuracy"] == 0.5

    y = np.array([0, 1])
    report = majority_sklearn_report(z, t, y)
    assert report["accuracy"] == 1


def test_majority_vote_random_guess():
    z = np.zeros((4, 4))
    t = np.zeros((4, 2))

    z[0, 0] = 1
    z[1, 1] = 1
    z[2, 0] = 1

    t[0, 0] = 1
    t[1, 1] = 1
    t[2, 0] = 1
    t[3, 0] = 1

    # no info about sample 4, thus random value is chosen
    # y_majority = np.array([0, 1, 0, random])
    y = np.array([0, 1, 1, 1])
    report = majority_sklearn_report(z, t, y)
    assert 0.5 <= report["accuracy"] <= 0.75



import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence
from torch.nn.utils.rnn import pad_packed_sequence


class BidirectionalLSTM(nn.Module):
    def __init__(
            self,
            word_input_dim,
            word_output_dim,
            word_embedding_matrix,
            num_classes,
            size_factor=200,
    ):
        super(BidirectionalLSTM, self).__init__()

        self.word_input_dim = word_input_dim
        self.word_output_dim = word_output_dim
        self.word_embedding_matrix = word_embedding_matrix
        self.size_factor = size_factor
        self.num_classes = num_classes

        self.word_embedding = nn.Embedding(
            word_input_dim, word_output_dim, padding_idx=0
        )
        self.word_embedding.weight = nn.Parameter(
            torch.tensor(word_embedding_matrix, dtype=torch.float32)
        )
        self.word_embedding.weight.requires_grad = False

        self.type_linear = nn.Linear(20, size_factor * 2)

        self.td_dense = nn.Linear(word_output_dim, size_factor)
        self.biLSTM = nn.LSTM(
            size_factor, size_factor, bidirectional=True, batch_first=True
        )

        self.predict = nn.Linear(size_factor * 2, num_classes)
        self.init_weights()

    def forward(self, x):
        word_embeddings = self.word_embedding(x)

        td_dense = self.td_dense(word_embeddings)

        biLSTM, (h_n, c_n) = self.biLSTM(td_dense)
        self.biLSTM.flatten_parameters()

        final_state = h_n.view(1, 2, x.shape[0], self.size_factor)[-1]
        h_1, h_2 = final_state[0], final_state[1]  # forward & backward pass
        concat = torch.cat((h_1, h_2), 1)  # concatenate both states
        final = self.predict(concat)

        return final

    def init_weights(self):
        """
        Here we reproduce Keras default initialization weights to initialize Embeddings/LSTM weights
        """
        torch.manual_seed(12345)
        ih = (
            param.data for name, param in self.named_parameters() if "weight_ih" in name
        )
        hh = (
            param.data for name, param in self.named_parameters() if "weight_hh" in name
        )
        b = (param.data for name, param in self.named_parameters() if "bias" in name)
        for t in ih:
            nn.init.xavier_uniform_(t)
        for t in hh:
            nn.init.orthogonal_(t)
        for t in b:
            nn.init.constant_(t, 0)
